"""
Module for handling molecules.
"""

import importlib.resources
import os
import re
from copy import deepcopy
from operator import itemgetter

import networkx as nx
import numpy as np
from monty.serialization import loadfn
from numpy.random import Generator
from pymatgen.core.structure import Molecule
from pymatgen.symmetry.analyzer import PointGroupAnalyzer, generate_full_symmops
from scipy.spatial.distance import cdist
from scipy.spatial.transform import Rotation

from pyxtal.constants import single_smiles
from pyxtal.database.collection import Collection
from pyxtal.database.element import Element
from pyxtal.msg import AtomTypeError, ConformerError
from pyxtal.operations import OperationAnalyzer, SymmOp, angle, rotate_vector
from pyxtal.symmetry import Group
from pyxtal.tolerance import Tol_matrix

with importlib.resources.as_file(importlib.resources.files("pyxtal") / "database" / "bonds.json") as path:
    bonds = loadfn(path)

molecule_collection = Collection("molecules")


def find_rotor_from_smile(smile):
    """
    Find the positions of rotatable bonds based on a SMILES string.

    Rotatable bonds are those which are not part of rings and which
    fit specific chemical patterns. These torsions are filtered by
    rules such as avoiding atoms with only one neighbor and avoiding
    equivalent torsions.

    Args:
        smile (str): The SMILES string representing the molecule.

    Returns:
        list of tuples: Each tuple represents a torsion as (i, j, k, l)
        where i-j-k-l are atom indices involved in the rotatable bond.

    """

    def cleaner(list_to_clean, neighbors):
        """
        Remove duplicate and invalid torsions from a list of atom index tuples.

        Filters torsions based on the neighbors count for the atoms involved in the torsion.
        This avoids torsions that involve terminal atoms and duplicates.

        Args:
            list_to_clean (list of tuples): List of torsions (i, j, k, l)
            neighbors (list of int): List of neighbors for each atom in the molecule.

        Returns:
            list of tuples: Cleaned list of torsions.
        """

        for_remove = []
        for x in reversed(range(len(list_to_clean))):
            ix0 = itemgetter(0)(list_to_clean[x])
            ix1 = itemgetter(0)(list_to_clean[x])
            ix2 = itemgetter(0)(list_to_clean[x])
            ix3 = itemgetter(3)(list_to_clean[x])
            # for i-j-k-l, we don't want i, l are the ending members
            # C-C-S=O is not a good choice since O is only 1-coordinated
            # C-C-NO2 is a good choice since O is only 1-coordinated

            # Remove torsions that involve terminal atoms with only one neighbor
            if neighbors[ix0] == 1 and neighbors[ix1] == 2 or neighbors[ix3] == 1 and neighbors[ix2] == 2:
                for_remove.append(x)
            else:
                # Remove duplicate torsions that are equivalent
                for y in reversed(range(x)):
                    ix1 = itemgetter(1)(list_to_clean[x])
                    ix2 = itemgetter(2)(list_to_clean[x])
                    iy1 = itemgetter(1)(list_to_clean[y])
                    iy2 = itemgetter(2)(list_to_clean[y])
                    if [ix1, ix2] == [iy1, iy2] or [ix1, ix2] == [iy2, iy1]:
                        for_remove.append(y)
        clean_list = []
        for i, v in enumerate(list_to_clean):
            if i not in set(for_remove):
                clean_list.append(v)
        return clean_list

    if smile in ["Cl-", "F-", "Br-", "I-", "Li+", "Na+"]:
        return []
    else:
        from rdkit import Chem

        # SMARTS patterns to identify rotatable bonds and double bonds
        smarts_torsion1 = "[*]~[!$(*#*)&!D1]-&!@[!$(*#*)&!D1]~[*]"
        smarts_torsion2 = "[*]~[^2]=[^2]~[*]"  # C=C bonds
        # smarts_torsion2="[*]~[^1]#[^1]~[*]" # C-C triples bonds, to be fixed

        mol = Chem.MolFromSmiles(smile)
        mol_with_H = Chem.AddHs(mol)
        N_atom = mol.GetNumAtoms()
        neighbors = [len(a.GetNeighbors()) for a in mol_with_H.GetAtoms()][:N_atom]
        # make sure that the ending members will be counted
        # neighbors[0] += 1; neighbors[-1] += 1
        patn_tor1 = Chem.MolFromSmarts(smarts_torsion1)
        torsion1 = cleaner(list(mol.GetSubstructMatches(patn_tor1)), neighbors)
        patn_tor2 = Chem.MolFromSmarts(smarts_torsion2)
        torsion2 = cleaner(list(mol.GetSubstructMatches(patn_tor2)), neighbors)

        # Combine and clean torsions
        tmp = cleaner(torsion1 + torsion2, neighbors)

        # Exclude torsions that are part of rings
        torsions = []
        for t in tmp:
            (i, j, k, l) = t
            b = mol.GetBondBetweenAtoms(j, k)
            if not b.IsInRing():
                torsions.append(t)
        # if len(torsions) > 6: torsions[1] = (4, 7, 10, 15)
        return torsions  # + [(6, 7, 8, 3), (6, 5, 4, 3)]


def has_non_aromatic_ring(smiles):
    """
    Determine if a molecule has a non-aromatic ring.
    It checks if a cyclic ring system exists that is not aromatic.

    Args:
        smiles (str): A SMILES string representing the molecule.

    Returns:
        bool: True if it contains a non-aromatic ring, False otherwise.
    """
    from rdkit import Chem

    # Convert the SMILES string to an RDKit molecule object
    mol = Chem.MolFromSmiles(smiles)

    # Check if the molecule has rings at all
    if not mol.HasSubstructMatch(Chem.MolFromSmarts("[R]")):
        return False  # No rings present

    # Get information about the rings in the molecule
    ring_info = mol.GetRingInfo()

    # Check each ring to see if it is aromatic; return True if a non-aromatic ring is found
    return any(
        not all(mol.GetBondWithIdx(idx).GetIsAromatic() for idx in ring) for ring in ring_info.BondRings()
    )  # No non-aromatic rings found


def generate_molecules(smile, wps=None, N_iter=5, N_conf=10, tol=0.5):
    """
    generate pyxtal_molecules from smiles codes.

    Args:
        smile: smiles code
        wps: list of wps
        N_iter: rdkit parameter
        N_conf: number of conformers
        tol: rdkit parameter

    Returns:
        a list of pyxtal molecules
    """

    from rdkit import Chem
    from rdkit.Chem import AllChem

    torsionlist = find_rotor_from_smile(smile)
    if len(torsionlist) == 0:
        if has_non_aromatic_ring(smile):
            Num = 10
    else:
        Num = len(tor)

    def get_conformers(smile, seed):
        mol = Chem.MolFromSmiles(smile)
        mol = Chem.AddHs(mol)
        ps = AllChem.ETKDGv3()
        ps.randomSeed = seed
        ps.runeRmsThresh = tol
        AllChem.EmbedMultipleConfs(mol, max([4, Num]), ps)
        return mol

    m0 = pyxtal_molecule(smile + ".smi", fix=True)
    _, valid = m0.get_orientations_in_wps(wps)
    mols = []
    if valid:
        # print('torsion', m0.get_torsion_angles())
        mols.append(m0)

    for i in range(N_iter):
        mol = get_conformers(smile, seed=i)
        res = AllChem.MMFFOptimizeMoleculeConfs(mol)
        for id, conf in enumerate(mol.GetConformers()):
            m = m0.copy()
            xyz = m.align(conf)
            m.reset_positions(xyz)
            m.get_symmetry(symmetrize=True)
            m.energy = res[id][1]
            _, valid = m.get_orientations_in_wps(wps)
            add = bool(valid)
            if add:
                match = False
                for mol in mols:
                    rms, _ = mol.get_rmsd2(xyz, mol.mol.cart_coords)
                    # print("rms", mol.get_torsion_angles(), rms)
                    if rms < tol:
                        match = True
                        break
                if not match:
                    # print(len(mols)+1, m.get_torsion_angles(xyz))
                    mols.append(m)
                    if len(mols) == N_conf:
                        return mols
    # for m in mols:
    #    print(m.energy, m.pga.sch_symbol, len(torsionlist))
    return mols


class pyxtal_molecule:
    """
    A molecule class to support the descriptin of molecules in a xtal

    Features:
    0. Parse the input from different formats (SMILES, xyz, gjf, etc.).
    1. Estimate molecular properties such as volume, tolerance, and radii.
    2. Find and store symmetry information of the molecule.
    3. Get the principal axis of the molecule.
    4. Re-align the molecule to center it at (0, 0, 0).

    SMILES Format:
    If a SMILES format is used, the molecular center is defined following
    RDKit's handling of molecular transformations:
    https://www.rdkit.org/docs/source/rdkit.Chem.rdMolTransforms.html

    Otherwise, the center is just the mean of atomic positions

    Args:
        mol (str or pymatgen.Molecule): The molecule representation, either as a string
            (SMILES or filename) or as a pymatgen `Molecule` object.
        tm (Tol_matrix, optional): A tolerance matrix object, used for molecular tolerances.
        symmetrize (bool, optional): Whether to symmetrize the molecule using its point group.
        fix (bool, optional): Fix torsions in the molecule.
        torsions (list, optional): List of torsions to analyze or fix.
        seed (int, optional): Random seed for internal processes. Defaults to a hex seed.
        random_state (int or numpy.Generator, optional): Numpy random state for random number generation.
        symtol (float, optional): Symmetry tolerance. Default is 0.3.
        active_sites (list, optional): List of active sites within the molecule.
    """

    def list_molecules():
        """
        list the internally supported molecules
        """
        molecule_collection.show_names()

    def __init__(
        self,
        mol=None,
        symmetrize=True,
        fix=False,
        torsions=None,
        seed=None,
        random_state=None,
        tm=Tol_matrix(prototype="molecular"),
        symtol=0.3,
        active_sites=None,
    ):

        mo = None
        self.smile = None
        self.torsionlist = [] #None
        self.reflect = False
        if seed is None:
            seed = 0xF00D
        self.seed = seed

        # Active sites is a two list of tuples [(donors), (acceptors)]
        self.active_sites = active_sites

        if isinstance(random_state, Generator):
            self.random_state = random_state.spawn(1)[0]
        else:
            self.random_state = np.random.default_rng(random_state)

        # Parse molecules: either file or molecule name
        if isinstance(mol, str):
            tmp = mol.split(".")
            self.name = tmp[0]
            if len(tmp) > 1:
                # Load the molecule from the given file
                if tmp[-1] in ["xyz", "gjf", "g03", "json"]:
                    if os.path.exists(mol):
                        mo = Molecule.from_file(mol)
                    else:
                        raise NameError(f"{mol:s} is not a valid path")
                elif tmp[-1] == "smi":
                    self.smile = tmp[0]
                    res = self.rdkit_mol_init(tmp[0], fix, torsions)
                    (symbols, xyz, self.torsionlist) = res
                    mo = Molecule(symbols, xyz)
                    symmetrize = False
                else:
                    raise NameError(f"{tmp[-1]:s} is not a supported format")
            else:
                # print('\nLoad the molecule {:s} from collections'.format(mol))
                mo = molecule_collection[mol]

        elif hasattr(mol, "sites"):  # pymatgen molecule
            self.name = str(mol.formula)
            mo = mol

        if mo is None:
            msg = f"Could not create molecules from given input: {mol:s}"
            raise NameError(msg)

        # Molecule and symmetry analysis
        self.props = mo.site_properties
        if len(mo) > 1 and symmetrize:
            try:
                pga = PointGroupAnalyzer(mo, symtol)
                mo = pga.symmetrize_molecule()["sym_mol"]
            except:
                print(
                    "Warning: Problem in parsing molecular symmetry with symtol=",
                    symtol,
                )
                print("Proceed with no symmetrization")
        self.mol = mo
        self.get_symmetry()

        # Additional molecular properties
        self.tm = tm
        self.box = self.get_box()
        self.volume = self.box.volume
        self.get_symbols()
        self.get_radius()
        self.tols_matrix = self.get_tols_matrix()
        xyz = self.mol.cart_coords
        self.reset_positions(xyz - self.get_center(xyz))
        if self.smile is not None and self.smile not in single_smiles:
            # print(self.smile)
            ori, _, self.reflect = self.get_orientation(xyz)

    def __str__(self):
        return "[" + self.name + "]"

    def save_str(self):
        """
        save the object as a dictionary
        """
        return self.mol.to(fmt="xyz")

    @classmethod
    def load_str(cls, string):
        """
        load the molecule from a dictionary
        """
        mol = Molecule.from_str(string, fmt="xyz")
        return cls(mol)

    def copy(self):
        """
        simply copy the structure
        """
        return deepcopy(self)

    def swap_axis(self, ax):
        """
        swap the molecular axis
        """
        coords = self.mol.cart_coords[:, ax]
        mo = Molecule(self.symbols, coords)

        return pyxtal_molecule(mo, self.tm)

    def get_box(self, padding=None):
        """
        Given a molecule, find a minimum orthorhombic box containing it.
        Size is calculated using min and max x, y, and z values,
        plus the padding defined by the vdw radius
        For best results, call oriented_molecule first.

        Args:
            padding: float (default is 3.4 according to the vdw radius of C)

        Returns:
            box: a Box object
        """
        mol, P = reoriented_molecule(self.mol)
        xyz = mol.cart_coords
        dims = [0, 0, 0]
        for i in range(3):
            dims[i] = np.max(xyz[:, i]) - np.min(xyz[:, i])
            if padding is not None:
                dims[i] += padding
                dims[i] = max([dims[i], 2.0])  # for planar molecules
            else:
                ids = np.argsort(xyz[:, i])
                r = Element(mol[ids[0]].species_string).vdw_radius
                r += Element(mol[ids[-1]].species_string).vdw_radius
                dims[i] = max([dims[i] + r, 3.4])  # for planar molecules
        return Box(dims)

    def get_lengths(self):
        if not hasattr(self, "box"):
            self.box = self.get_box()
        return self.box.width, self.box.height, self.box.length

    def get_max_length(self):
        w, h, l = self.get_lengths()
        return max([w, h, l])

    def get_box_coordinates(self, xyz, padding=0, resolution=1.0):
        """
        create the points cloud to describe the molecular box

        Args:
            center: molecular center position
            orientation: orientation matrix
            padding: padding of the box
            resolution: float in angstrom

        Return:
            cell: box axis
            vertices: [N, 3] array, Cartesian coordinates to describe the box.
            center: box center
        """
        cell = self.get_principle_axes(xyz).T
        center = self.get_center(xyz)  # , geometry=True)
        box = self.get_box(padding)
        # print(box)
        w, h, l = box.width, box.height, box.length
        cell[0, :] *= l
        cell[1, :] *= w
        cell[2, :] *= h
        x_ = np.linspace(-1 / 2, 1 / 2, int(l / resolution) + 1)
        y_ = np.linspace(-1 / 2, 1 / 2, int(w / resolution) + 1)
        z_ = np.linspace(-1 / 2, 1 / 2, int(h / resolution) + 1)

        # XY
        # print(len(x_), len(y_), len(z_))
        x, y = np.meshgrid(x_, y_, indexing="ij")
        size = len(x.flatten())
        xy = np.zeros([size * 2, 3])
        xy[:size, 0] = x.flatten()
        xy[size:, 0] = x.flatten()
        xy[:size, 1] = y.flatten()
        xy[size:, 1] = y.flatten()
        xy[:size, 2] = -0.5
        xy[size:, 2] = 0.5
        # print(xy.shape)
        # print(xy)

        y, z = np.meshgrid(y_, z_, indexing="ij")
        size = len(y.flatten())
        yz = np.zeros([size * 2, 3])
        yz[:size, 1] = y.flatten()
        yz[size:, 1] = y.flatten()
        yz[:size, 2] = z.flatten()
        yz[size:, 2] = z.flatten()
        yz[:size, 0] = -0.5
        yz[size:, 0] = 0.5

        x, z = np.meshgrid(x_, z_, indexing="ij")
        size = len(z.flatten())
        xz = np.zeros([size * 2, 3])
        xz[:size, 0] = x.flatten()
        xz[size:, 0] = x.flatten()
        xz[:size, 2] = z.flatten()
        xz[size:, 2] = z.flatten()
        xz[:size, 1] = -0.5
        xz[size:, 1] = 0.5

        vertices = np.zeros([len(xy) + len(yz) + len(xz), 3])
        vertices[: len(xy), :] = xy
        vertices[len(xy) : len(xy) + len(yz), :] = yz
        vertices[len(xy) + len(yz) :, :] = xz
        vertices = vertices.dot(cell)
        vertices += center

        return cell, vertices, center

    def get_radius(self):
        """
        get the radius of a molecule
        """
        r_max = 0
        for coord, number in zip(self.mol.cart_coords, self.mol.atomic_numbers):
            radius = np.linalg.norm(coord) + 0.5 * self.tm.get_tol(number, number)
            if radius > r_max:
                r_max = radius
        self.radius = r_max
        # reestimate the radius if it has stick shape
        rmax = max([self.box.width, self.box.height, self.box.length])
        rmin = min([self.box.width, self.box.height, self.box.length])
        if rmax / rmin > 3 and rmax > 12:
            self.radius = rmin

    def get_symbols(self):
        self.symbols = [specie.name for specie in self.mol.species]

    def get_tols_matrix(self, mol2=None, tm=None):
        """
        Compute the 2D tolerance matrix between the current and other molecules

        Args:
            mol2: the 2nd pyxtal_molecule object
            tm: tolerance class

        Returns:
            a 2D matrix which is used internally for distance checking.
        """
        if tm is None:
            tm = self.tm

        numbers1 = self.mol.atomic_numbers
        numbers2 = self.mol.atomic_numbers if mol2 is None else mol2.mol.atomic_numbers

        tols = np.zeros((len(numbers1), len(numbers2)))
        for i1, number1 in enumerate(numbers1):
            for i2, number2 in enumerate(numbers2):
                tols[i1, i2] = tm.get_tol(number1, number2)
                # allow hydrogen bond
                if [number1, number2] in [
                    [1, 7],
                    [1, 8],
                    [1, 9],
                    [7, 1],
                    [8, 1],
                    [9, 1],
                ]:
                    tols[i1, i2] *= 0.9

        if len(self.mol) == 1:
            tols *= 0.8  # if only one atom, reduce the tolerance
        return tols

    def set_labels(self):
        """
        Set atom labels for the given molecule for H-bond caculation.
        Needs to identify the following:
        - (O)N-H
        - acid-O
        - amide-O
        - alcohol-O
        - N with NH2
        - N with NH
        """

        def search_H(pairs, ref, pos_H):
            """
            quick routine to search for the id of H that is bonded to N/O:

            Args:
                pairs: list of atomic pairs
                ref: reference id for O or N
                pos_H: the starting position for H
            """
            res = []
            for p in pairs:
                if ref in p and max(p) >= pos_H:
                    res.append(max(p))
            return res

        if len(self.mol) > 1:
            from rdkit import Chem

            # template
            acid1 = Chem.MolFromSmarts("[C,c]C(=O)O")  # COOH
            acid2 = Chem.MolFromSmarts("[CH](=O)O")  # COOH
            amide1 = Chem.MolFromSmarts("[C,c]C(=O)N")  # CONH
            amide2 = Chem.MolFromSmarts("[CH](=O)N")  # CONH
            alcohol = Chem.MolFromSmarts("[c,CX3][OH]")  # ROH
            # alcohol2 = Chem.MolFromSmarts('c[OH]')  #ROH
            Chem.MolFromSmarts("c")  # Aromatic
            NH1 = Chem.MolFromSmarts("[NH1]")  # NH1
            NH2 = Chem.MolFromSmarts("[NH2]")  # NH2

            # Initialize mol
            m = Chem.MolFromSmiles(self.smile)
            pos_H = m.GetNumAtoms()  # starting position for H
            m = Chem.AddHs(m)
            labels = [a.GetSymbol() for a in m.GetAtoms()]

            # Create bonds
            bonds = m.GetBonds()
            pairs = np.zeros([len(bonds), 2], dtype=int)
            for i, bond in enumerate(bonds):
                pairs[i, 0] = bond.GetBeginAtomIdx()
                pairs[i, 1] = bond.GetEndAtomIdx()

            # Assign aromatic
            # ds = m.GetSubstructMatches(aromatic_carbon)
            # for d in ds: labels[d[0]] += '_aromatic'

            # Assign O
            N_O = labels.count("O")
            if N_O > 0:
                count_O = 0
                for i, smart in enumerate([acid1, acid2, amide1, amide2, alcohol]):
                    ds = m.GetSubstructMatches(smart)
                    # print(i, ds)
                    for d in ds:
                        if i in [0, 1]:  # COOH or COO in general
                            if i == 0:
                                labels[d[2]] += "_acid"
                                id = 3
                                # labels[d[3]] += '_acid'
                            else:
                                id = 2
                                labels[d[1]] += "_acid"

                            Hs = search_H(pairs, d[id], pos_H)
                            if len(Hs) > 0:
                                labels[Hs[0]] += "_O"
                            count_O += 2

                        elif i in [2, 3]:  # CONH
                            id = 3 if i == 2 else 2
                            labels[d[id - 1]] += "_amide"
                            count_O += 1

                        else:  # OH
                            labels[d[-1]] += "_alcohol"
                            labels[search_H(pairs, d[-1], pos_H)[0]] += "_O"
                            count_O += 1
                    if count_O == N_O:
                        # print(i, count_O, 'break')
                        break

            # Assign N
            N_N = labels.count("N")
            if N_N > 0:
                count_N = 0
                for i, smart in enumerate([NH1, NH2]):
                    ds = m.GetSubstructMatches(smart)
                    for d in ds:
                        if i == 0:
                            labels[d[0]] += "_H1"  # N_H2
                            labels[search_H(pairs, d[0], pos_H)[0]] += "_N"
                        else:
                            labels[d[0]] += "_H2"  # N_H2
                            Hs = search_H(pairs, d[0], pos_H)
                            labels[Hs[0]] += "_N"
                            labels[Hs[1]] += "_N"
                        count_N += 1
                    if count_N == N_N:
                        break
        else:
            labels = self.symbols
            print(labels)
        self.labels = labels

    def get_coefs_matrix(self, mol2=None, ignore_error=True):
        """
        Get the Atom-Atom potential parameters
        E = A*exp(-B*R) - C*R^(-6)
        according to Gavezotti, Acc. Chem. Res., 27, 1994
        in Kcal/mol and angstrom

        Args:
            mol2: the 2nd pyxtal_molecule object
            tm: tolerance class

        Returns:
            a 3D matrix for computing the intermolecular energy
        """
        labels1 = self.labels if hasattr(self, "labels") else self.symbols
        numbers1 = self.mol.atomic_numbers
        if mol2 is None:
            numbers2 = self.mol.atomic_numbers
            labels2 = labels1
        else:
            numbers2 = mol2.mol.atomic_numbers
            labels2 = mol2.labels if hasattr(mol2, "labels") else mol2.symbols

        coefs = np.zeros([len(numbers1), len(numbers2), 3])
        for i1, n1 in enumerate(numbers1):
            for i2, n2 in enumerate(numbers2):
                if [n1, n2] in [[1, 1]]:  # H-H
                    coefs[i1, i2, :] = [5774, 4.01, 26.1]
                elif [n1, n2] in [[1, 6], [6, 1]]:  # H-C
                    coefs[i1, i2, :] = [28870, 4.10, 113.0]
                elif [n1, n2] in [[1, 7]]:  # H-N
                    if len(labels1[i1]) > 1:
                        if labels1[i1] == "H_N1":  # HB-N(-NH-N):
                            coefs[i1, i2, :] = [7215600, 7.78, 476]
                        else:  # HB-N(-NH2-N):
                            coefs[i1, i2, :] = 1803920, 7.37, 165
                    else:
                        coefs[i1, i2, :] = [54560, 4.52, 120.0]
                elif [n1, n2] in [[7, 1]]:  # N-H
                    if len(labels2[i2]) > 1:
                        if labels2[i2] == "H_N1":  # HB-N(-NH-N):
                            coefs[i1, i2, :] = [7215600, 7.78, 476]
                        else:  # HB-N(-NH2-N):
                            coefs[i1, i2, :] = 1803920, 7.37, 165
                    else:
                        coefs[i1, i2, :] = [54560, 4.52, 120.0]
                elif [n1, n2] in [[1, 8]]:  # H-O
                    if len(labels1[i1]) > 1:
                        if labels2[i2] == "O_amide":  # HB...O=C-N
                            coefs[i1, i2, :] = [3607810, 7.78, 238]
                        elif labels2[i2] == "O_acid":  # HB...O=C-OH
                            coefs[i1, i2, :] = [6313670, 8.75, 205]
                        elif labels2[i2] == "O_alcohol":  # HB...OH
                            coefs[i1, i2, :] = [4509750, 7.78, 298]
                        else:
                            # print("Oxygen label problem", labels2[i2]); import sys; sys.exit()
                            coefs[i1, i2, :] = [70610, 4.82, 105.0]
                    else:  # Normal cases:
                        coefs[i1, i2, :] = [70610, 4.82, 105.0]

                elif [n1, n2] in [[8, 1]]:  # O-H
                    if len(labels2[i2]) > 1:
                        if labels1[i1] == "O_amide":  # HB...O=C-N
                            coefs[i1, i2, :] = [3607810, 7.78, 238]
                        elif labels1[i1] == "O_acid":  # HB...O=C-OH
                            coefs[i1, i2, :] = [6313670, 8.75, 205]
                        elif labels1[i1] == "O_alcohol":  # HB...OH
                            coefs[i1, i2, :] = [4509750, 7.78, 298]
                        else:
                            # print('Oxygen label problem', labels2[i1]); import sys; sys.exit()
                            coefs[i1, i2, :] = [70610, 4.82, 105.0]
                    else:  # Normal cases:
                        coefs[i1, i2, :] = [70610, 4.82, 105.0]

                elif [n1, n2] in [[1, 16], [16, 1]]:  # H-S
                    coefs[i1, i2, :] = [64190, 4.03, 279.0]

                elif [n1, n2] in [[1, 17], [17, 1]]:  # H-Cl
                    coefs[i1, i2, :] = [70020, 4.09, 279.0]

                elif [n1, n2] in [[6, 6]]:  # C-C
                    coefs[i1, i2, :] = [54050, 3.47, 578.0]

                elif [n1, n2] in [[6, 7], [7, 6]]:  # C-N
                    coefs[i1, i2, :] = [117470, 3.86, 667.0]

                elif [n1, n2] in [[6, 8], [8, 6]]:  # C-O
                    coefs[i1, i2, :] = [93950, 3.74, 641.0]

                elif [n1, n2] in [[6, 16], [16, 6]]:  # C-S
                    coefs[i1, i2, :] = [126460, 3.41, 1504.0]

                elif [n1, n2] in [[6, 17], [17, 6]]:  # C-Cl
                    coefs[i1, i2, :] = [93370, 3.52, 923.0]

                elif [n1, n2] in [[7, 7]]:  # N-N
                    coefs[i1, i2, :] = [87300, 3.65, 691.0]

                elif [n1, n2] in [[7, 8], [8, 7]]:  # N-O
                    coefs[i1, i2, :] = [64190, 3.86, 364.0]

                elif [n1, n2] in [[7, 16], [16, 7], [7, 17], [17, 7]]:  # N-S/Cl
                    coefs[i1, i2, :] = [0, 3.65, 0]

                elif [n1, n2] in [[8, 8]]:  # O-O
                    if False:  # labels1[i1] == 'O_alcohol' and labels2[i2] == 'O_alcohol':
                        coefs[i1, i2, :] = [3607800, 5.00, 3372.0]
                    else:
                        coefs[i1, i2, :] = [46680, 3.74, 319.0]

                elif [n1, n2] in [[8, 16], [16, 8]]:  # O-S
                    coefs[i1, i2, :] = [110160, 3.63, 906.0]

                elif [n1, n2] in [[8, 17], [17, 8]]:  # O-Cl
                    coefs[i1, i2, :] = [80855, 3.63, 665.0]

                elif [n1, n2] in [[16, 16]]:  # S-S
                    coefs[i1, i2, :] = [259960, 3.52, 2571]

                elif [n1, n2] in [[16, 17], [17, 16]]:  # S-Cl
                    coefs[i1, i2, :] = [1800000, 3.52, 2000]  # made

                elif [n1, n2] in [[17, 17]]:  # Cl-Cl
                    coefs[i1, i2, :] = [140050, 3.52, 1385]
                else:
                    if ignore_error:
                        coefs[i1, i2, :] = [0.0, 0.0, 0.0]
                    else:
                        msg = f"atom type is not supported: {n1:d} {n2:d}"
                        raise AtomTypeError(msg)
                        # return None
        return coefs

    def show(self):
        """
        show the molecule
        """
        from pyxtal.viz import display_molecules

        return display_molecules([self.mol])

    def show_box(self, center=np.zeros(3), orientation=None):
        """
        show the molecule
        """
        from pyxtal.viz import display_molecules

        return display_molecules(self.mol)

    def rdkit_mol(self, N_confs=1):
        """
        initialize the mol xyz and torsion list
        """
        from rdkit import Chem

        mol = Chem.MolFromMolBlock(self.rdkit_mb, removeHs=False)
        if N_confs > 1:
            conf = mol.GetConformer(0)
            for _i in range(N_confs - 1):
                mol.AddConformer(conf, True)
        return mol

    def rdkit_mol_init(self, smile, fix, torsions):
        """
        initialize the mol xyz and torsion list

        Args:
            smile: smile string
            fix: whether or not fix the seed
            torsions: None or list
        """
        from rdkit import Chem
        from rdkit.Chem import AllChem

        if smile not in single_smiles:  # ["Cl-", "F-", "Br-", "I-", "Li+", "Na+"]:
            torsionlist = find_rotor_from_smile(smile)
            mol = Chem.MolFromSmiles(smile)
            mol = Chem.AddHs(mol)
            symbols = []
            for id in range(mol.GetNumAtoms()):
                symbols.append(mol.GetAtomWithIdx(id).GetSymbol())
            if len(smile) > 100:  # a tmp fix for KEKULN10
                AllChem.EmbedMultipleConfs(mol, numConfs=1, randomSeed=3)
                cid = 0
            else:
                ps = AllChem.ETKDGv3()
                ps.randomSeed = self.seed
                AllChem.EmbedMultipleConfs(mol, 1, ps)
                if mol.GetNumConformers() == 0:
                    AllChem.EmbedMultipleConfs(mol, 3, ps)

            res = AllChem.MMFFOptimizeMoleculeConfs(mol)
            engs = [c[1] for c in res]
            cid = engs.index(min(engs))
            self.rdkit_mb = Chem.MolToMolBlock(mol)
            self.energy = engs[cid]
            ref_conf = mol.GetConformer(cid)  # always the reference molecule

            if fix or torsions is not None or len(torsionlist) == 0:
                conf = ref_conf
            else:
                randomSeed = -1 if self.random_state is None else 1
                AllChem.EmbedMultipleConfs(
                    mol,
                    numConfs=max([1, 4 * len(torsionlist)]),
                    maxAttempts=200,
                    useRandomCoords=True,
                    pruneRmsThresh=0.5,
                    randomSeed=randomSeed,
                )
                N_confs = mol.GetNumConformers()
                conf_id = int(self.random_state.choice(range(N_confs)))
                conf = mol.GetConformer(conf_id)
                # xyz = conf.GetPositions()
                # res = AllChem.MMFFOptimizeMoleculeConfs(mol)
                # print("Eng", res[conf_id])

            # set tosion angles from random or pre-defined values
            if torsions is not None:
                xyz = self.set_torsion_angles(conf, torsions, torsionlist=torsionlist)
            else:
                xyz = conf.GetPositions()
                xyz -= self.get_center(xyz)
        else:
            # single atom cation or anions
            pattern = r"[A-Za-z]+(?=[+\-]?[^A-Za-z]|$)"
            matches = re.findall(pattern, smile)
            if matches:
                symbols = [matches[0]]  # ["Cl"]
            else:
                raise ValueError("the input smiles cannot be analyzed", smile)
            xyz = np.zeros([1, 3])
            torsionlist = []
        return symbols, xyz, torsionlist

    def perturb_torsion(self, xyz):
        """
        slightly perturb the torsion
        """
        angs = self.get_torsion_angles(xyz, self.torsionlist)
        angs *= 1 + 0.1 * self.random_state.uniform(-1.0, 1.0, len(angs))
        xyz = self.set_torsion_angles(conf, angs, torsionlist=self.torsionlist)
        xyz -= self.get_center(xyz)
        return xyz

    def align(self, conf, reflect=False, torsionlist=None):
        """
        Align the molecule and return the xyz
        The default CanonicalizeConformer function may also include inversion
        """
        from rdkit.Chem import rdMolTransforms as rdmt

        # Rotation
        if len(self.smile) > 1:
            trans = rdmt.ComputeCanonicalTransform(conf)
            if abs(abs(np.linalg.det(trans)) - 1.0) > 1e-1:
                print("Bug in trans", np.linalg.det(trans))
                import sys

                sys.exit()
            elif np.linalg.det(trans[:3, :3]) < 0:
                trans[:3, :3] *= -1

            # add reflection if needed
            if reflect:
                trans[:3, :3] *= -1
            rdmt.TransformConformer(conf, trans)

        # Translation
        pt = rdmt.ComputeCentroid(conf)
        center = np.array([pt.x, pt.y, pt.z])
        xyz = conf.GetPositions() - center

        # adjust cases like H2O
        if len(self.smile) == 1:
            xyz -= np.mean(xyz, axis=0)
            A = get_inertia_tensor(xyz)
            P = np.linalg.eigh(A)[1]
            if np.linalg.det(P) < 0:
                P[0] *= -1
            xyz = np.dot(xyz, P)
        return xyz

    def get_center(self, xyz, geometry=False):
        """
        get the molecular center for a transformed xyz
        """
        if geometry or self.smile is None:
            return np.mean(xyz, axis=0)
        else:
            if self.smile in [
                "Cl-",
                "F-",
                "Br-",
                "I-",
                "Li+",
                "Na+",
                "[Cl-]",
                "[F-]",
                "[Br-]",
                "[I-]",
                "[Li+]",
                "[Na+]",
            ]:
                return xyz[0]
            else:
                if len(self.smile) == 1:
                    # return xyz[0]
                    return np.mean(xyz, axis=0)
                else:
                    # from rdkit
                    from rdkit.Chem import rdMolTransforms as rdmt
                    from rdkit.Geometry import Point3D

                    conf = self.rdkit_mol().GetConformer(0)
                    for i in range(len(xyz)):
                        x, y, z = xyz[i]
                        conf.SetAtomPosition(i, Point3D(x, y, z))
                    pt = rdmt.ComputeCentroid(conf)

                    return np.array([pt.x, pt.y, pt.z])

    def get_principle_axes(self, xyz, rdmt=True):
        """
        get the principle axis for a rotated xyz, sorted by the moments
        """
        if self.smile is None or len(self.smile) == 1 or not rdmt:
            Inertia = get_inertia_tensor(xyz)
            _, matrix = np.linalg.eigh(Inertia)
            return matrix
        else:
            from rdkit.Chem import rdMolTransforms as rdmt
            from rdkit.Geometry import Point3D

            conf1 = self.rdkit_mol().GetConformer(0)
            for i in range(len(self.mol)):
                x, y, z = xyz[i]
                conf1.SetAtomPosition(i, Point3D(x, y, z))

            return rdmt.ComputePrincipalAxesAndMoments(conf1)[0]

    def get_torsion_angles(self, xyz=None, torsionlist=None):
        """
        get the torsion angles
        """
        if xyz is None:
            xyz = self.mol.cart_coords
        if torsionlist is None:
            torsionlist = self.torsionlist

        angs = []
        if len(torsionlist) > 0:

            from rdkit.Chem import rdMolTransforms as rdmt
            from rdkit.Geometry import Point3D

            conf = self.rdkit_mol().GetConformer(0)
            for i in range(len(xyz)):
                x, y, z = xyz[i]
                conf.SetAtomPosition(i, Point3D(x, y, z))

            for torsion in torsionlist:
                (i, j, k, l) = torsion
                angs.append(rdmt.GetDihedralDeg(conf, i, j, k, l))
        return angs

    def set_torsion_angles(self, conf, angles, reflect=False, torsionlist=None):
        """
        reset the torsion angles and update molecular xyz
        """
        from rdkit.Chem import rdMolTransforms as rdmt

        if torsionlist is None:
            torsionlist = self.torsionlist
        for id, torsion in enumerate(torsionlist):
            (i, j, k, l) = torsion
            rdmt.SetDihedralDeg(conf, i, j, k, l, angles[id])

        return self.align(conf, reflect)

    def relax(self, xyz, align=False):
        """
        Relax the input xyz coordinates with rdit MMFF

        Args:
            xyz: 3D coordinates
            align: whether or not align the xyz

        Returns:
            xyz: new xyz
            eng: energy value
        """

        from rdkit.Chem import AllChem
        from rdkit.Geometry import Point3D

        mol = self.rdkit_mol(1)
        conf0 = mol.GetConformer(0)
        # reset the xyz
        for i in range(len(self.mol)):
            x, y, z = xyz[i]
            conf0.SetAtomPosition(i, Point3D(x, y, z))
        res = AllChem.MMFFOptimizeMoleculeConfs(mol)
        xyz = self.align(conf0) if align else mol.GetConformer(0).GetPositions()
        return xyz, res[0][1]

    def get_rmsd2(self, xyz0, xyz1):
        """
        Compute the rmsd with another 3D xyz coordinates

        Args:
            xyz: 3D coordinates

        Returns:
            rmsd:
            transition matrix:
        """

        from rdkit.Chem import RemoveHs, rdMolAlign
        from rdkit.Geometry import Point3D

        mol = self.rdkit_mol(3)
        conf0 = mol.GetConformer(0)
        conf1 = mol.GetConformer(1)
        for i in range(len(self.mol)):
            x0, y0, z0 = xyz0[i]
            x1, y1, z1 = xyz1[i]
            conf0.SetAtomPosition(i, Point3D(x0, y0, z0))
            conf1.SetAtomPosition(i, Point3D(x1, y1, z1))

        mol = RemoveHs(mol)
        rmsd, trans = rdMolAlign.GetAlignmentTransform(mol, mol, 1, 0)

        return rmsd, trans

    def get_rmsd(self, xyz, debug=False):
        """
        Compute the rmsd with another 3D xyz coordinates

        Args:
            xyz: 3D coordinates

        Returns:
            rmsd (float): rmsd values
            transition matrix: 4*4 matrix
            match: True or False
        """

        from rdkit.Chem import RemoveHs, rdMolAlign
        from rdkit.Geometry import Point3D

        mol = self.rdkit_mol(3)
        # 3 conformers for comparison
        conf0 = mol.GetConformer(0)
        conf1 = mol.GetConformer(1)  # reference+reflection
        conf2 = mol.GetConformer(2)  # trial xyz
        angs = self.get_torsion_angles(xyz)
        xyz0 = self.set_torsion_angles(conf0, angs)  # conf0 with aligned
        xyz1 = self.set_torsion_angles(conf0, angs, True)  # conf0 with aligned+reflect
        # print('xyz0', xyz0)
        # reset the xyz
        for i in range(len(self.mol)):
            x0, y0, z0 = xyz0[i]
            x1, y1, z1 = xyz1[i]
            x, y, z = xyz[i]
            conf0.SetAtomPosition(i, Point3D(x0, y0, z0))
            conf1.SetAtomPosition(i, Point3D(x1, y1, z1))
            conf2.SetAtomPosition(i, Point3D(x, y, z))

        mol = RemoveHs(mol)
        rmsd1, trans1 = rdMolAlign.GetAlignmentTransform(mol, mol, 2, 0)
        rmsd2, trans2 = rdMolAlign.GetAlignmentTransform(mol, mol, 2, 1)

        if debug:
            from rdkit.Chem import rdmolfiles

            rdmolfiles.MolToXYZFile(mol, "1.xyz", 0)
            rdmolfiles.MolToXYZFile(mol, "2.xyz", 1)
            rdmolfiles.MolToXYZFile(mol, "3.xyz", 2)
            print(rmsd1, rmsd2)
        if rmsd1 <= rmsd2:
            # return rmsd1, trans1, True
            return rmsd1, trans1, False
        else:
            # return rmsd2, trans2, False
            return rmsd2, trans2, True

    def get_orientation(self, xyz, rtol=0.15):
        """
        For the given xyz, compute the orientation

        Args:
            xyz: molecular xyz
        """

        xyz -= self.get_center(xyz)

        if self.smile is not None and len(self.smile) > 1:  # not in ["O", "o"]:
            rmsd, trans, reflect = self.get_rmsd(xyz)
            tol = rtol * len(xyz)

            if rmsd < tol:
                trans = trans[:3, :3].T
                r = Rotation.from_matrix(trans)
                return r.as_euler("zxy", degrees=True), rmsd, reflect
            else:
                msg = "Problem in conformer\n"
                msg += f"{rmsd1:5.2f} {rmsd2:5.2f}\n"
                if len(self.torsionlist) > 0:
                    msg += str(self.get_torsion_angles(xyz)) + "\n"
                    msg += str(self.get_torsion_angles(xyz0)) + "\n"
                    msg += str(self.get_torsion_angles(xyz1)) + "\n"
                raise ConformerError(msg)
        else:
            # the orientation of CH4, NH3, H2O
            ref = np.array(
                [
                    [-0.00111384, 0.36313718, 0.0],
                    [-0.82498189, -0.18196256, 0.0],
                    [0.82609573, -0.18117463, 0.0],
                ]
            )
            Inertia = get_inertia_tensor(xyz)
            _, matrix = np.linalg.eigh(Inertia)

            np.dot(xyz, matrix)
            # identify the rotation matrix
            libs = np.array(
                [
                    [[1, 1, 1]],
                    [[-1, 1, 1]],
                    [[1, -1, 1]],
                    [[1, 1, -1]],
                    [[-1, -1, 1]],
                    [[1, -1, -1]],
                    [[-1, 1, -1]],
                    [[-1, -1, -1]],
                ]
            )
            dists = np.zeros(8)
            for i, lib in enumerate(libs):
                matrix0 = matrix * np.repeat(lib, 3, axis=0)
                res = np.dot(ref, np.linalg.inv(matrix0))
                dists[i] = np.sum((res - xyz[:len(ref)]) ** 2)
                # print(i, res)
            id = np.argmin(dists)
            matrix = matrix * np.repeat(libs[id], 3, axis=0)

            r = Rotation.from_matrix(np.linalg.inv(matrix).T)
            ang = r.as_euler("zxy", degrees=True)
            return ang, 0, False

    def to_ase(self):
        """
        Convert to ase atoms
        """
        return self.mol.to_ase_atoms()

    def reset_positions(self, coors):
        """
        reset the coordinates
        """
        from pymatgen.core.sites import Site

        if len(coors) != len(self.mol._sites):
            raise ValueError("number of atoms is inconsistent!")
        else:
            for i, coor in enumerate(coors):
                _site = self.mol._sites[i]
                new_site = Site(_site.species, coor, properties=_site.properties)
                self.mol._sites[i] = new_site

    def apply_inversion(self):
        """
        reset the coordinates
        """

        xyz = self.mol.cart_coords
        center = self.get_center(xyz)
        xyz -= center
        xyz *= -1
        self.reset_positions(xyz)

    def get_symmetry(self, xyz=None, symmetrize=False, rtol=0.30):
        """
        Set the molecule's point symmetry.
            - pga: pymatgen.symmetry.analyzer.PointGroupAnalyzer object
            - pg: pyxtal.symmetry.Group object
            - symops: a list of SymmOp objects

        Args:
            symmetrize: boolean, whether or not symmetrize the coordinates
        """
        mol = deepcopy(self.mol) if xyz is None else Molecule(self.symbols, xyz)

        if self.smile is not None:
            mol.remove_species("H")
            mol._spin_multiplicity = None  # don't check spin

        if symmetrize:
            pga = PointGroupAnalyzer(mol, rtol, eigen_tolerance=1e-3)
            mol = pga.symmetrize_molecule()["sym_mol"]
        pga = PointGroupAnalyzer(mol, rtol, eigen_tolerance=1e-3)
        self.mol_no_h = mol

        # print(mol.to(fmt='xyz'), pga.sch_symbol)
        # For single atoms, no point group using a list of operations
        if len(mol) == 1:
            symm_m = []
            symbol = "C1"
        else:
            symbol = pga.sch_symbol
            pg = pga.get_pointgroup()
            symm_m = list(pg)

            if "*" in symbol:  # linear molecules
                symbol = symbol.replace("*", "6")
                # Add 12-fold  and reflections in place of ininitesimal rotation
                for i, axis in enumerate(np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])):
                    # op = SymmOp.from_rotation_and_translation(aa2matrix(axis, np.pi/6), [0,0,0])
                    m1 = Rotation.from_rotvec(np.pi / 6 * axis).as_matrix()
                    op = SymmOp.from_rotation_and_translation(m1, [0, 0, 0])
                    if pga.is_valid_op(op):
                        symm_m.append(op)
                        # Any molecule with infinitesimal symmetry is linear;
                        # Thus, it possess mirror symmetry for any axis perpendicular
                        # To the rotational axis. pymatgen does not add this symmetry
                        # for all linear molecules - for example, hydrogen
                        if i == 0:
                            symm_m.append(SymmOp.from_xyz_str("x,-y,z"))
                            symm_m.append(SymmOp.from_xyz_str("x,y,-z"))
                            # r = SymmOp.from_xyz_str("-x,y,-z")
                        elif i == 1:
                            symm_m.append(SymmOp.from_xyz_str("-x,y,z"))
                            symm_m.append(SymmOp.from_xyz_str("x,y,-z"))
                            # r = SymmOp.from_xyz_str("-x,-y,z")
                        elif i == 2:
                            symm_m.append(SymmOp.from_xyz_str("-x,y,z"))
                            symm_m.append(SymmOp.from_xyz_str("x,-y,z"))
                            # r = SymmOp.from_xyz_str("x,-y,-z")
                        # Generate a full list of SymmOps for the pointgroup
                        symm_m = generate_full_symmops(symm_m, 1e-3)
                        break
        self.symops = symm_m
        self.pga = pga
        if symbol == "S2":
            symbol = "Ci"
        self.pg = Group(symbol, dim=0)

    def get_orientations_in_wps(self, wps=None, rtol=1e-2):
        """
        Compute the valid orientations from a given Wyckoff site symmetry.

        Args:
            wp: a pyxtal.symmetry.Wyckoff_position object
        Returns:
            a list of operations.Orientation objects
        """
        if wps is None:
            return None, True
        else:
            valid = False
            valid_oris = []
            for wp in wps:
                allowed = self.get_orientations_in_wp(wp, rtol)
                if len(allowed) > 0:
                    valid = True
                valid_oris.append(allowed)
            return valid_oris, valid

    def get_orientations_in_wp(self, wp, rtol=1e-2):
        """
        Compute the valid orientations from a given Wyckoff site symmetry.

        Args:
            wp: a pyxtal.symmetry.Wyckoff_position object
        Returns:
            a list of pyxtal.molecule.Orientation objects
        """
        # For single atoms, there are no constraints
        if len(self.mol) == 1 or wp.index == 0:
            return [Orientation([[1, 0, 0], [0, 1, 0], [0, 0, 1]], degrees=2, random_state=self.random_state)]
        # C1 molecule cannot take specical position
        elif wp.index > 1 and self.pga.sch_symbol == "C1":
            return []

        symm_w = wp.get_site_symm_wo_translation()  # symmetry without translation
        # molecule has fewer symops
        if len(self.pg[0]) < len(symm_w):
            return []

        symm_m = self.symops
        opa_m = []
        for op_m in symm_m:
            opa = OperationAnalyzer(op_m)
            opa_m.append(opa)

        # Store OperationAnalyzer objects for each Wyckoff symmetry SymmOp
        opa_w = []
        for op_w in symm_w:
            opa_w.append(OperationAnalyzer(op_w))

        """
        Check for constraints from the Wyckoff symmetry. If we find ANY two
        constraints (SymmOps with unique axes), the molecule's point group MUST
        contain SymmOps which can be aligned to these particular constraints.
        However, there may be multiple compatible orientations of the molecule
        consistent with these constraints
        """
        constraint1 = None
        constraint2 = None
        for i, op_w in enumerate(symm_w):
            if opa_w[i].axis is not None:
                constraint1 = opa_w[i]
                for j, op_w in enumerate(symm_w):
                    if opa_w[j].axis is not None:
                        dot = np.dot(opa_w[i].axis, opa_w[j].axis)
                        if (not np.isclose(dot, 1, rtol=rtol)) and (not np.isclose(dot, -1, rtol=rtol)):
                            constraint2 = opa_w[j]
                            break
                break
        # Indirectly store the angle between the constraint axes
        if constraint1 is not None and constraint2 is not None:
            dot_w = np.dot(constraint1.axis, constraint2.axis)
        # Generate 1st consistent molecular constraints
        constraints_m = []
        if constraint1 is not None:
            for i, opa1 in enumerate(opa_m):
                if opa1.is_conjugate(constraint1):
                    constraints_m.append([opa1, []])
                    # Generate 2nd constraint in opposite direction
                    extra = deepcopy(opa1)
                    extra.axis = [
                        opa1.axis[0] * -1,
                        opa1.axis[1] * -1,
                        opa1.axis[2] * -1,
                    ]
                    constraints_m.append([extra, []])

        # Remove redundancy for the first constraints
        list_i = list(range(len(constraints_m)))
        list_j = list(range(len(constraints_m)))
        copy = deepcopy(constraints_m)
        for i, c1 in enumerate(copy):
            if i in list_i:
                for j, c2 in enumerate(copy):
                    if i > j and j in list_j and j in list_i:
                        # Check if axes are colinear
                        if np.isclose(np.dot(c1[0].axis, c2[0].axis), 1, rtol=rtol):
                            list_i.remove(j)
                            list_j.remove(j)
                        # Check if axes are symmetrically equivalent
                        else:
                            cond1 = False
                            # cond2 = False
                            for opa in opa_m:
                                if opa.type == "rotation":
                                    op = opa.op
                                    if np.isclose(
                                        np.dot(op.operate(c1[0].axis), c2[0].axis),
                                        1,
                                        rtol=5 * rtol,
                                    ):
                                        cond1 = True
                                        break
                            if cond1:  # or cond2 is True:
                                list_i.remove(j)
                                list_j.remove(j)
        c_m = deepcopy(constraints_m)
        constraints_m = []
        for i in list_i:
            constraints_m.append(c_m[i])

        # Generate 2nd consistent molecular constraints
        valid = list(range(len(constraints_m)))
        if constraint2 is not None:
            for i, c in enumerate(constraints_m):
                opa1 = c[0]
                for j, opa2 in enumerate(opa_m):
                    if opa2.is_conjugate(constraint2):
                        dot_m = np.dot(opa1.axis, opa2.axis)
                        # Ensure that the angles are equal
                        if abs(dot_m - dot_w) < 0.02 or abs(dot_m + dot_w) < 0.02:
                            constraints_m[i][1].append(opa2)
                            # Generate 2nd constraint in opposite direction
                            extra = deepcopy(opa2)
                            extra.axis = [
                                opa2.axis[0] * -1,
                                opa2.axis[1] * -1,
                                opa2.axis[2] * -1,
                            ]
                            constraints_m[i][1].append(extra)
                # If no consistent constraints are found, remove first constraint
                if constraints_m[i][1] == []:
                    valid.remove(i)
        copy = deepcopy(constraints_m)
        constraints_m = []
        for i in valid:
            constraints_m.append(copy[i])

        # Generate orientations consistent with the possible constraints
        orientations = []
        # Loop over molecular constraint sets
        for c1 in constraints_m:
            v1 = c1[0].axis
            v2 = constraint1.axis
            T = rotate_vector(v1, v2)
            # If there is only one constraint
            if c1[1] == []:
                o = Orientation(T, degrees=1, axis=constraint1.axis, random_state=self.random_state)
                orientations.append(o)
            else:
                # Loop over second molecular constraints
                for opa in c1[1]:
                    phi = angle(constraint1.axis, constraint2.axis)
                    phi2 = angle(constraint1.axis, np.dot(T, opa.axis))
                    if np.isclose(phi, phi2, rtol=rtol):
                        r = np.sin(phi)
                        c = np.linalg.norm(np.dot(T, opa.axis) - constraint2.axis)
                        theta = np.arccos(1 - (c**2) / (2 * (r**2)))
                        # R = aa2matrix(constraint1.axis, theta)
                        R = Rotation.from_rotvec(theta * constraint1.axis).as_matrix()
                        T2 = np.dot(R, T)
                        a = angle(np.dot(T2, opa.axis), constraint2.axis)
                        if not np.isclose(a, 0, rtol=rtol):
                            T2 = np.dot(np.linalg.inv(R), T)
                        o = Orientation(T2, degrees=0, random_state=self.random_state)
                        orientations.append(o)

        # Ensure the identity orientation is checked if no constraints are found
        if constraints_m == []:
            o = Orientation(np.identity(3), degrees=2, random_state=self.random_state)
            orientations.append(o)
        # Remove redundancy from orientations
        list_i = list(range(len(orientations)))
        list_j = list(range(len(orientations)))
        for i, o1 in enumerate(orientations):
            if i in list_i:
                for j, o2 in enumerate(orientations):
                    if i > j and j in list_j and j in list_i:
                        # m1 = o1.get_matrix(angle=0)
                        # m2 = o2.get_matrix(angle=0)
                        m1 = o1.matrix
                        m2 = o2.matrix
                        new_op = SymmOp.from_rotation_and_translation(np.dot(m2, np.linalg.inv(m1)), [0, 0, 0])
                        P = SymmOp.from_rotation_and_translation(np.linalg.inv(m1), [0, 0, 0])
                        old_op = P * new_op * P.inverse
                        if self.pga.is_valid_op(old_op):
                            list_i.remove(j)
                            list_j.remove(j)
        orientations_new = []
        for i in list_i:
            orientations_new.append(orientations[i])

        # Check each of the found orientations for consistency with Wyckoff site.
        # If consistent, put into an array of valid orientations
        # print("======", orientations_new)
        allowed = []
        for o in orientations_new:
            op = o.get_op()
            mo = deepcopy(self.mol_no_h)
            mo.apply_operation(op)
            # print(mo)
            if is_compatible_symmetry(mo, wp):
                allowed.append(o)
        return allowed

    def get_energy(self, xyz1, xyz2):
        """
        Get packing energy between two neighboring molecules
        """
        cdist(xyz1 - xyz2)


class Box:
    """
    Class for storing the binding box for a molecule.

    Args:
        dims: [length, width, height]
    """

    def __init__(self, dims):
        self.length = dims[0]  # float(abs(maxy - miny))
        self.width = dims[1]  # float(abs(maxx - minx))
        self.height = dims[2]  # float(abs(maxz - minz))
        self.volume = self.width * self.length * self.height

    def __str__(self):
        return f"l: {self.length:6.2f}, w: {self.width:6.2f}, d: {self.height:6.2f}"

    def operate(self, rot=np.eye(3), center=np.zeros(3)):
        """
        Perform operation on the box:

        Args:
            rot: 3*3 rotation matrix
            center: center position
        """
        raise NotImplementedError


class Orientation:
    """
    Stores orientations for molecules based on vector constraints.
    Can be stored to regenerate orientations consistent with a given constraint
    vector, without re-calling orientation_in_wyckoff_position. Allows for
    generating orientations which differ only in their rotation about a given
    axis.

    Args:
        matrix: a 3x3 rotation matrix describing the orientation (and/or
            inversion) to store
        degrees: the number of degrees of freedom...
            0 - The orientation refers to a single rotation matrix
            1 - The orientation can be rotated about a single axis
            2 - The orientation can be any pure rotation matrix

        axis:
            an optional axis about which the orientation can rotate. Only used
            if degrees is equal to 1
    """

    def __init__(self, matrix=None, degrees=2, axis=None, random_state=None):
        self.matrix = np.array(matrix)
        self.degrees = degrees

        if isinstance(random_state, Generator):
            self.random_state = random_state.spawn(1)[0]
        else:
            self.random_state = np.random.default_rng(random_state)

        if degrees == 1:
            if axis is None:
                raise ValueError("axis is required for orientation")

            axis /= np.linalg.norm(axis)
        self.axis = axis

        self.r = Rotation.from_matrix(self.matrix)
        self.angle = None

    def __str__(self):
        s = "-------PyXtal.molecule.Orientation class----\n"
        s += f"degree of freedom: {self.degrees:d}\n"
        s += "Rotation matrix:\n"
        s += "{:6.3f} {:6.3f} {:6.3f}\n".format(*self.matrix[:, 0])
        s += "{:6.3f} {:6.3f} {:6.3f}\n".format(*self.matrix[:, 1])
        s += "{:6.3f} {:6.3f} {:6.3f}\n".format(*self.matrix[:, 2])
        if self.axis is not None:
            s += "Rotation axis\n"
            s += "{:6.2f} {:6.2f} {:6.3f}\n".format(*self.axis)
        return s

    def reset_matrix(self, matrix):
        self.matrix = matrix
        self.r = Rotation.from_matrix(matrix)

    def __repr__(self):
        return str(self)

    def copy(self):
        return deepcopy(self)

    def save_dict(self):
        return {"matrix": self.matrix, "degrees": self.degrees, "axis": self.axis}

    @classmethod
    def load_dict(cls, dicts):
        matrix = dicts["matrix"]
        degrees = dicts["degrees"]
        axis = dicts["axis"]
        return cls(matrix, degrees, axis)

    def change_orientation(self, angle="random", flip=False, update=True):
        """
        Change the orientation of molecule by applying a rotation.

        It allows for specification of an angle (or a random angle) to rotate about
        the constraint axis. If the system has 2 degrees of rotational freedom,
        the molecule can also be flipped with a probability

        Args:
            angle (float or str, optional): The angle to rotate about the constraint axis.
                                        If "random", a random rotation angle is selected
            flip (bool, optional): Whether to apply an random flip. This is only applied
                               if the system has 2 degrees of rotational freedom.
        """
        if self.degrees >= 1:
            # Choose the axis
            if self.axis is None:
                self.set_axis()

            # Parse the angle
            if angle == "random":
                angle = (self.random_state.random() - 1) * np.pi * 2
            #self.angle = angle

            # Update the matrix
            r1 = Rotation.from_rotvec(angle * self.axis)

            # Optionally flip the molecule
            if self.degrees == 2 and flip and self.random_state.random() > 0.5:
                ax = self.random_state.choice(["x", "y", "z"])
                angle0 = self.random_state.choice([90, 180, 270])
                r2 = Rotation.from_euler(ax, angle0, degrees=True)
                r1 = r2 * r1

            r = r1 * self.r
            matrix = r.as_matrix()
            if update:
                self.r = r
                self.matrix = matrix
            return matrix
        else:
            return self.matrix


    def set_axis(self):
        if self.degrees == 2:
            axis = self.random_state.random(3) - 0.5
            self.axis = axis / np.linalg.norm(axis)
            self.angle = 0

    def rotate_by_matrix(self, matrix, ignore_constraint=True):
        """
        rotate

        Args:
            matrix: 3*3 rotation matrix

        """
        if not ignore_constraint:
            if self.degrees == 0:
                raise ValueError("cannot rotate")

            if self.degrees == 1:
                axis = self.axis
                vec = Rotation.from_matrix(matrix).as_rotvec()
                if angle(vec, self.axis) > 1e-2 and angle(vec, -self.axis) > 1e-2:
                    raise ValueError("must rotate along the given axis")
        else:
            axis = None

        matrix = matrix.dot(self.matrix)
        return Orientation(matrix, self.degrees, axis, self.random_state)

    def get_matrix(self, angle="random"):
        """
        Generate a 3x3 rotation matrix consistent with the orientation's
        constraints. Allows for specification of an angle (possibly random) to
        rotate about the constraint axis.

        Args:
            angle: an angle to rotate about the constraint axis. If "random",
                chooses a random rotation angle. If self.degrees==2, chooses a
                random 3d rotation matrix to multiply by. If the original matrix
                is wanted, set angle=0, or call self.matrix

        Returns:
            a 3x3 rotation (and/or inversion) matrix (numpy array)
        """
        if self.degrees == 2:
            if angle == "random":
                axis = self.random_state.sample(3)
                axis = axis / np.linalg.norm(axis)
                angle = self.random_state.random() * np.pi * 2
            else:
                axis = self.axis
            return Rotation.from_rotvec(angle * axis).as_matrix()

        elif self.degrees == 1:
            angle = self.random_state.random() * np.pi * 2 if angle == "random" else self.angle
            return Rotation.from_rotvec(angle * self.axis).as_matrix()

        elif self.degrees == 0:
            return self.matrix
        return None

    def get_op(self):  # , angle=None):
        """
        Generate a SymmOp object consistent with the orientation's constraints.
        Allows for specification of an angle (possibly random) to rotate about
        the constraint axis.

        Args:
            angle: an angle to rotate about the constraint axis. If "random",
                chooses a random rotation angle. If self.degrees==2, chooses a
                random 3d rotation matrix to multiply by. If the original matrix
                is wanted, set angle=0, or call self.matrix

        Returns:
            pymatgen.core.structure. SymmOp object
        """
        # if angle is not None:
        #    self.change_orientation(angle)
        return SymmOp.from_rotation_and_translation(self.matrix, [0, 0, 0])

    def random_orientation(self):
        """
        Applies random rotation (if possible) and returns a new orientation with
        the new base matrix.

        Returns:
            a new orientation object with a different base rotation matrix
        """

        self.change_orientation()
        return self

    def get_Euler_angles(self):
        """
        get the Euler angles
        """
        return self.r.as_euler("zxy", degrees=True)


def get_inertia_tensor(coords, weights=None):
    """
    Calculate the symmetric inertia tensor for a molecule.

    Args:
        coords: [N, 3] array of coordinates

    Returns:
        a 3x3 numpy array representing the inertia tensor
    """
    if weights is None:
        weights = np.ones(len(coords))
    coords -= np.mean(coords, axis=0)
    Inertia = np.zeros([3, 3])
    Inertia[0, 0] = np.sum(weights * coords[:, 1] ** 2 + weights * coords[:, 2] ** 2)
    Inertia[1, 1] = np.sum(weights * coords[:, 0] ** 2 + weights * coords[:, 2] ** 2)
    Inertia[2, 2] = np.sum(weights * coords[:, 0] ** 2 + weights * coords[:, 1] ** 2)
    Inertia[0, 1] = Inertia[1, 0] = -np.sum(weights * coords[:, 0] * coords[:, 1])
    Inertia[0, 2] = Inertia[2, 0] = -np.sum(weights * coords[:, 0] * coords[:, 2])
    Inertia[1, 2] = Inertia[2, 1] = -np.sum(weights * coords[:, 1] * coords[:, 2])

    return Inertia


def reoriented_molecule(mol):  # , nested=False):
    """
    Allign a molecule so that its principal axes is the identity matrix.

    Args:
        mol: a Molecule object

    Returns:
        new_mol: a reoriented copy of the original molecule.
        P: the 3x3 rotation matrix used to obtain it.
    """
    coords = mol.cart_coords
    numbers = mol.atomic_numbers
    coords -= np.mean(coords, axis=0)
    A = get_inertia_tensor(coords)
    # Store the eigenvectors of the inertia tensor
    P = np.linalg.eigh(A)[1]
    if np.linalg.det(P) < 0:
        P[:, 0] *= -1
    coords = np.dot(coords, P)
    return Molecule(numbers, coords), P


def is_compatible_symmetry(mol, wp):
    """
    Tests if a molecule meets the symmetry requirements of a Wyckoff position

    Args:
        mol: a pymatgen Molecule object.
        wp: a pyxtal.symmetry.Wyckoff_position object
    """
    # For single atoms, there are no constraints
    if len(mol) == 1 or wp.index == 0:
        return True
    pga = PointGroupAnalyzer(mol)
    return all(pga.is_valid_op(op) for op in wp.get_site_symm_wo_translation())


def make_graph(mol, tol=0.2):
    """
    make graph object for the input molecule
    """
    # print("making graphs")
    G = nx.Graph()
    names = {}
    for i, site in enumerate(mol._sites):
        names[i] = site.specie.value
        if names[i] not in ["C", "H", "O", "N", "S", "P", "Si", "F", "Cl", "Br", "I"]:
            raise ValueError(f"{names[i]} is not supported")

    for i in range(len(mol) - 1):
        site1 = mol.sites[i]
        for j in range(i + 1, len(mol)):
            site2 = mol.sites[j]
            key = f"{names[i]:s}-{names[j]:s}"
            if site1.distance(site2) < bonds[key]:
                G.add_edge(i, j)
                # print(key, site1.distance(site2))
    nx.set_node_attributes(G, names, "name")

    return G


def compare_mol_connectivity(mol1, mol2, ignore_name=False):
    """
    Compare two molecules by connectivity
    """

    G1 = make_graph(mol1)
    G2 = make_graph(mol2)
    if ignore_name:
        GM = nx.isomorphism.GraphMatcher(G1, G2)
    else:
        fun = lambda n1, n2: n1["name"] == n2["name"]
        GM = nx.isomorphism.GraphMatcher(G1, G2, node_match=fun)

    return GM.is_isomorphic(), GM.mapping


if __name__ == "__main__":
    smiles = "CCN1C(=O)c2ccc3C(=O)N(CC)C(=O)c4ccc(C1=O)c2c34"
    ans1 = [(0, 1, 2, 20), (13, 12, 11, 14)]
    print(ans1)
    ans2 = find_rotor_from_smile(smiles)
    print(ans2)
    assert ans1 == ans2

    smiles = "Nc1c(Cl)cc(cc1N(=O)=O)N(=O)=O"
    ans2 = find_rotor_from_smile(smiles)
    print(ans2)
    ans1 = [(6, 5, 11, 13), (6, 7, 8, 10)]
    print(ans1)
    assert ans1 == ans2

    smiles = "COc1cc(C=O)ccc1O"
    ans2 = find_rotor_from_smile(smiles)
    print(ans2)
    ans1 = [(0, 1, 2, 9), (6, 5, 4, 7)]
    print(ans1)
    assert ans1 == ans2



"""ASE LAMMPS Calculator Library Version"""

import importlib.resources
import os

import ase.units
import numpy as np
from ase.calculators.calculator import Calculator
from ase.constraints import UnitCellFilter, FixSymmetry
from ase.optimize import LBFGS
from ase.optimize.fire import FIRE
from lammps import lammps


#
def opt_lammpslib(
    struc,
    lmp,
    parameters=None,
    mask=None,
    logfile="-",
    path="tmp",
    calc_type=None,
    lmp_file=None,
    molecule=False,
    method="FIRE",
    fmax=0.01,
    opt_cell=False,
    a=0.1,
    steps=500,
):
    """
    mask: [1, 1, 1, 1, 1, 1], a, b, c, alpha, beta, gamma
          [1, 1, 0, 0, 0, 1]
    """

    if lmp_file is not None:
        lammps = LAMMPSlib(
            lmp=lmp,
            lmp_file=lmp_file,
            log_file="lammps.log",
            molecule=molecule,
            path=path,
        )
    elif calc_type is not None:
        lammps = LAMMPSlib(
            lmp=lmp,
            lmpcmds=parameters,
            calc_type=calc_type,
            log_file="lammps.log",
            molecule=molecule,
            path=path,
        )
    else:
        lammps = LAMMPSlib(
            lmp=lmp,
            lmpcmds=parameters,
            log_file="lammps.log",
            molecule=molecule,
            path=path,
        )

    # check_symmetry(si, 1.0e-6, verbose=True)
    struc.set_calculator(lammps)
    struc.set_constraint(FixSymmetry(struc))
    if opt_cell:
        ecf = UnitCellFilter(struc, mask)
        dyn = FIRE(ecf, logfile=logfile, a=a) if method == "FIRE" else LBFGS(ecf, logfile=logfile)
    else:
        dyn = FIRE(struc, logfile=logfile) if method == "FIRE" else LBFGS(struc, logfile=logfile)
    dyn.run(fmax=fmax, steps=steps)
    return struc


def run_lammpslib(
    struc,
    lmp,
    parameters=None,
    path="tmp",
    calc_type=None,
    lmp_file=None,
    molecule=False,
    method="opt",
    temp=300,
    steps=10000,
    comp_z=None,
    min_style="cg",
):
    if lmp_file is not None:
        lammps = LAMMPSlib(
            lmp=lmp,
            lmp_file=lmp_file,
            log_file="lammps.log",
            molecule=molecule,
            path=path,
        )

    elif calc_type is not None:
        lammps = LAMMPSlib(
            lmp=lmp,
            lmpcmds=parameters,
            calc_type=calc_type,
            log_file="lammps.log",
            molecule=molecule,
            path=path,
        )
    else:
        parameter0 = [*parameters, "neighbor 1.0 bin", "neigh_modify  every 1  delay 1  check yes"]

        if method == "md":
            parameter0 += [
                "fix 1 all nvt temp " + str(temp) + " " + str(temp) + " 0.05",
                "timestep  0.001",
                "thermo 1000",
                "run " + str(steps),
                "reset_timestep 0",
                f"min_style {min_style}",
                f"minimize 1e-15 1e-15 {steps} {steps}",
                "thermo 0",
            ]
        elif method == "opt":
            if comp_z is not None:
                parameter0 += [f"change_box all z scale {comp_z}"]
            parameter0 += [
                "timestep  0.001",
                f"min_style {min_style}",
                f"minimize 1e-15 1e-15 {steps} {steps}",
            ]

        else:
            parameter0 += ["run 0"]

        lammps = LAMMPSlib(
            lmp=lmp,
            lmpcmds=parameter0,
            molecule=molecule,
            lmp_file=lmp_file,
            log_file="lammps.log",
            path=path,
        )

    struc.set_calculator(lammps)

    Eng = struc.get_potential_energy()

    return struc, Eng


def is_upper_triangular(arr, atol=1e-8):
    """test for upper triangular matrix based on numpy"""
    # must be (n x n) matrix
    assert len(arr.shape) == 2
    assert arr.shape[0] == arr.shape[1]
    return np.allclose(np.tril(arr, k=-1), 0.0, atol=atol) and np.all(np.diag(arr) >= 0.0)


def convert_cell(ase_cell):
    """
    Convert a parallelepiped (forming right hand basis)
    to lower triangular matrix LAMMPS can accept. This
    function transposes cell matrix so the bases are column vectors
    """
    cell = ase_cell.T
    if not is_upper_triangular(cell):
        tri_mat = np.zeros((3, 3))
        A = cell[:, 0]
        B = cell[:, 1]
        C = cell[:, 2]
        tri_mat[0, 0] = np.linalg.norm(A)
        Ahat = A / np.linalg.norm(A)
        AxBhat = np.cross(A, B) / np.linalg.norm(np.cross(A, B))
        tri_mat[0, 1] = np.dot(B, Ahat)
        tri_mat[1, 1] = np.linalg.norm(np.cross(Ahat, B))
        tri_mat[0, 2] = np.dot(C, Ahat)
        tri_mat[1, 2] = np.dot(C, np.cross(AxBhat, Ahat))
        tri_mat[2, 2] = np.linalg.norm(np.dot(C, AxBhat))
        # create and save the transformation for coordinates
        volume = np.linalg.det(ase_cell)
        trans = np.array([np.cross(B, C), np.cross(C, A), np.cross(A, B)])
        trans /= volume
        coord_transform = np.dot(tri_mat, trans)
        return tri_mat, coord_transform
    else:
        return cell, None


class LAMMPSlib(Calculator):
    def __init__(
        self,
        lmp,
        lmpcmds=None,
        path="tmp",
        molecule=False,
        lmp_file=None,
        calc_type="single",
        *args,
        **kwargs,
    ):
        Calculator.__init__(self, *args, **kwargs)
        self.lmp = lmp
        self.lmp_file = lmp_file
        self.molecule = molecule
        self.calc_type = calc_type
        self.folder = path
        if not os.path.exists(self.folder):
            os.makedirs(self.folder)
        self.lammps_data = self.folder + "/data.lammps"
        self.lammps_in = self.folder + "/in.lammps"
        with importlib.resources.as_file(importlib.resources.files("pyxtal") / "potentials") as path:
            self.ffpath = str(path)
        self.lmpcmds = lmpcmds
        self.paras = []
        if lmpcmds is not None:
            for para in lmpcmds:
                if not self.molecule and para.find("pair_coeff") >= 0:
                    tmp = para.split()
                    filename = tmp[3]
                    filename1 = self.ffpath + "/" + filename
                    para = para.replace(filename, filename1)
                self.paras.append(para)

    def calculate(self, atoms):
        """
        prepare lammps .in file and data file
        write_lammps_data(filename, self.atoms, )
        """
        boundary = ""
        for i in range(3):
            if atoms.pbc[i]:
                boundary += "p "
            else:
                boundary += "f "
        if boundary in ["f f p ", "p p f "]:  # needs some work later
            boundary = "p p p "
        self.boundary = boundary
        if self.molecule:
            self.write_lammps_data_water(atoms)
        else:
            self.write_lammps_data(atoms)
        self.write_lammps_in()
        self.lmp.file(self.lammps_in)
        # Extract the forces and energy
        self.lmp.command("variable pxx equal pxx")
        self.lmp.command("variable pyy equal pyy")
        self.lmp.command("variable pzz equal pzz")
        self.lmp.command("variable pxy equal pxy")
        self.lmp.command("variable pxz equal pxz")
        self.lmp.command("variable pyz equal pyz")
        self.lmp.command("variable fx atom fx")
        self.lmp.command("variable fy atom fy")
        self.lmp.command("variable fz atom fz")
        self.lmp.command("variable pe equal pe")
        if self.calc_type.find("GB") >= 0:
            self.lmp.command("variable Etot equal c_eatoms")
            self.gb_energy = self.lmp.extract_variable("Etot", None, 0)
            # print('gb_energy from lammps: ', self.gb_energy)
            # print('update lammps GB energy')

        pos = np.array(list(self.lmp.gather_atoms("x", 1, 3))).reshape(-1, 3)

        self.energy = self.lmp.extract_variable("pe", None, 0)
        # print('update lammps energy')

        xlo = self.lmp.extract_global("boxxlo", 1)
        xhi = self.lmp.extract_global("boxxhi", 1)
        ylo = self.lmp.extract_global("boxylo", 1)
        yhi = self.lmp.extract_global("boxyhi", 1)
        zlo = self.lmp.extract_global("boxzlo", 1)
        zhi = self.lmp.extract_global("boxzhi", 1)
        xy = self.lmp.extract_global("xy", 1)
        yz = self.lmp.extract_global("yz", 1)
        xz = self.lmp.extract_global("xz", 1)
        unitcell = np.array([[xhi - xlo, xy, xz], [0, yhi - ylo, yz], [0, 0, zhi - zlo]]).T

        stress = np.empty(6)
        stress_vars = ["pxx", "pyy", "pzz", "pyz", "pxz", "pxy"]

        for i, var in enumerate(stress_vars):
            stress[i] = self.lmp.extract_variable(var, None, 0)
        # print('update lammps stress')

        stress_mat = np.zeros((3, 3))
        stress_mat[0, 0] = stress[0]
        stress_mat[1, 1] = stress[1]
        stress_mat[2, 2] = stress[2]
        stress_mat[1, 2] = stress[3]
        stress_mat[2, 1] = stress[3]
        stress_mat[0, 2] = stress[4]
        stress_mat[2, 0] = stress[4]
        stress_mat[0, 1] = stress[5]
        stress_mat[1, 0] = stress[5]
        stress[0] = stress_mat[0, 0]
        stress[1] = stress_mat[1, 1]
        stress[2] = stress_mat[2, 2]
        stress[3] = stress_mat[1, 2]
        stress[4] = stress_mat[0, 2]
        stress[5] = stress_mat[0, 1]

        self.stress = -stress * 1e5 * ase.units.Pascal
        f = np.array(self.lmp.gather_atoms("f", 1, 3)).reshape(-1, 3) * (ase.units.eV / ase.units.Angstrom)
        # print('update lammps force')
        self.forces = f.copy()
        atoms.positions = pos.copy()
        atoms.cell = unitcell.copy()
        if self.molecule:
            atoms.positions *= 0.529
            atoms.cell *= 0.529
        self.atoms = atoms.copy()
        # self.atoms.info['GB_energy'] = self.gb_energy
        # print('update lammps all')

    def write_lammps_in(self):
        if self.lmp_file is not None:
            # print('cp ' + self.lmp_file + ' ' + self.folder + '/in.lammps')
            os.system("cp " + self.lmp_file + " " + self.folder + "/in.lammps")
        else:
            with open(self.lammps_in, "w") as fh:
                fh.write("clear\n")
                fh.write("box  tilt large\n")
                if self.molecule:
                    fh.write("units electron\n")
                else:
                    fh.write("units metal\n")
                if self.calc_type.find("GB") >= 0:
                    fh.write("boundary p p s\n")
                else:
                    fh.write(f"boundary {self.boundary:s}\n")
                fh.write("atom_modify sort 0 0.0\n")
                if not self.molecule:
                    fh.write(f"read_data {self.lammps_data:s}\n")
                fh.write("\n### interactions\n")
                for para in self.paras:
                    fh.write(f"{para:s}\n")
                fh.write("neighbor 1.0 bin\n")
                fh.write("neigh_modify  every 1  delay 1  check yes\n")
                if self.calc_type.find("GB") >= 0:
                    tmp = LAMMPS_collections().dict["GB_group"]
                    for i in tmp:
                        fh.write(i)
                if self.calc_type.find("GB_opt") >= 0:
                    tmp = LAMMPS_collections().dict["GB_opt"]
                    for i in tmp:
                        fh.write(i)

                elif self.calc_type.find("GB_md") >= 0:
                    tmp = LAMMPS_collections().dict["GB_opt"]
                    tmp += LAMMPS_collections(temp=1000).dict["GB_md"]
                    tmp += LAMMPS_collections().dict["GB_opt"]
                    for i in tmp:
                        fh.write(i)

                if self.calc_type.find("GB") >= 0:
                    fh.write("thermo_style custom pe pxx pyy pzz pyz pxz pxy c_eatoms\n")
                else:
                    fh.write("thermo_style custom pe pxx pyy pzz pyz pxz pxy\n")
                fh.write("thermo_modify flush yes\n")
                fh.write("thermo 1\n")
                fh.write("run 1\n")

                # fh.write('print "__end_of_ase_invoked_calculation__"\n')

    def write_lammps_data(self, atoms):
        atom_types = np.array(atoms.get_tags())  # [1]*len(atoms)
        if sum(atom_types) == 0:
            atom_types = np.array([1] * len(atoms))
        with open(self.lammps_data, "w") as fh:
            comment = "lammpslib autogenerated data file"
            fh.write(comment.strip() + "\n\n")
            fh.write(f"{len(atoms):d} atoms\n")
            fh.write(f"{len(np.unique(atom_types)):d} atom types\n")
            cell, coord_transform = convert_cell(atoms.get_cell())
            # cell = atoms.get_cell()
            fh.write("\n")
            fh.write(f"{0.0:16.8e} {cell[0, 0]:16.8e} xlo xhi\n")
            fh.write(f"{0.0:16.8e} {cell[1, 1]:16.8e} ylo yhi\n")
            fh.write(f"{0.0:16.8e} {cell[2, 2]:16.8e} zlo zhi\n")
            fh.write(f"{cell[0, 1]:16.8e} {cell[0, 2]:16.8e} {cell[1, 2]:16.8e} xy xz yz\n")

            fh.write("\n\nAtoms \n\n")
            for i, (typ, pos) in enumerate(zip(atom_types, atoms.get_positions())):
                if coord_transform is not None:
                    pos = np.dot(coord_transform, pos.transpose())
                fh.write(f"{i + 1:4d} {typ:4d} {pos[0]:16.8e} {pos[1]:16.8e} {pos[2]:16.8e}\n")

    def write_lammps_data_water(self, atoms):
        """
        Lammps input only for water model
        """
        [1] * len(atoms)
        N_atom = len(atoms)
        N_mol = int(len(atoms) / 3)
        N_bond = N_mol * 2
        N_angle = N_mol
        np.unique(atoms.numbers)
        lmp_types = np.zeros(N_atom, dtype=int)
        lmp_types[atoms.numbers == 1] = 2
        lmp_types[atoms.numbers == 8] = 1

        mol_types = np.zeros(N_atom, dtype=int)
        for i in range(N_mol):
            mol_types[i * 3 : (i + 1) * 3] = i + 1

        with open(self.lammps_data, "w") as fh:
            comment = "lammpslib autogenerated data file"
            fh.write(comment.strip() + "\n\n")
            fh.write(f"{N_atom} atoms\n")
            fh.write(f"{N_bond} bonds\n")
            fh.write(f"{N_angle} angles\n")

            fh.write("\n2 atom types\n")
            fh.write("1 bond types\n")
            fh.write("1 angle types\n")

            # cell = atoms.get_cell()/0.529
            cell, coord_transform = convert_cell(atoms.get_cell())
            cell /= 0.529
            fh.write("\n")
            fh.write(f"{0.0:16.8e} {cell[0, 0]:16.8e} xlo xhi\n")
            fh.write(f"{0.0:16.8e} {cell[1, 1]:16.8e} ylo yhi\n")
            fh.write(f"{0.0:16.8e} {cell[2, 2]:16.8e} zlo zhi\n")
            fh.write(f"{cell[0, 1]:16.8e} {cell[0, 2]:16.8e} {cell[1, 2]:16.8e} xy xz yz\n")

            fh.write("\n\nMasses \n\n")
            fh.write("  1 15.9994\n")
            fh.write("  2  1.0000\n")

            fh.write("\n\nBond Coeffs \n\n")
            fh.write(" 1    1.78    0.2708585 -0.327738785 0.231328959\n")

            fh.write("\n\nAngle Coeffs \n\n")
            fh.write("  1    0.0700  107.400000")
            fh.write("\n\nAtoms \n\n")
            for i, (typ, mtyp, pos) in enumerate(zip(lmp_types, mol_types, atoms.get_positions() / 0.529)):
                if coord_transform is not None:
                    pos = np.dot(coord_transform, pos.transpose())
                # print(i, mtyp, typ)
                if typ == 2:
                    fh.write(f"{i + 1:4d} {mtyp:4d} {typ:4d}   0.5564 {pos[0]:16.8f} {pos[1]:16.8f} {pos[2]:16.8f}\n")
                else:
                    fh.write(f"{i + 1:4d} {mtyp:4d} {typ:4d}  -1.1128 {pos[0]:16.8f} {pos[1]:16.8f} {pos[2]:16.8f}\n")

            fh.write("\nBonds \n\n")
            for i in range(N_mol):
                fh.write(f"{i * 2 + 1:4d} {1:4d} {i * 3 + 1:4d} {i * 3 + 2:4d}\n")
                fh.write(f"{i * 2 + 2:4d} {1:4d} {i * 3 + 1:4d} {i * 3 + 3:4d}\n")

            fh.write("\nAngles \n\n")
            for i in range(N_angle):
                fh.write(f"{i + 1:4d} {1:4d} {i * 3 + 2:4d} {i * 3 + 1:4d} {i * 3 + 3:4d}\n")

    def update(self, atoms):
        if not hasattr(self, "atoms") or self.atoms != atoms:
            self.calculate(atoms)

    def get_potential_energy(self, atoms, force_consistent=False):
        self.update(atoms)
        if self.calc_type.find("GB") >= 0:
            return self.gb_energy
        else:
            return self.energy

    def get_forces(self, atoms):
        self.update(atoms)
        return self.forces.copy()

    def get_stress(self, atoms):
        self.update(atoms)
        return self.stress.copy()


class LAMMPS_collections:
    """
    A class to provide the lammps input file
    """

    def __init__(self, temp=500):
        self.dict = {}

        self.dict["GB_group"] = [
            "# ---------GB--------\n",
            "group lowlayer type 1\n",
            "group gbregion type 2\n",
            "group uplayer type 3\n",
            "#----Define compute quantities-----\n",
            "compute eng all pe/atom \n",
            "compute eatoms gbregion reduce sum c_eng\n",
        ]

        self.dict["GB_opt"] = [
            "# ---------- Run GB Minimization -----\n",
            "fix f1 lowlayer setforce 0.0 0.0 0.0\n",
            "fix f2 uplayer aveforce 0.0 0.0 0.0\n",
            "min_style cg\n",
            "#dump   dump_all all custom 100000 dump.xyz element x y z\n",
            "minimize 1e-15 1e-15 10000 10000\n",
            "run 0\n",
            "thermo 0\n",
            "thermo_style custom pe pxx c_eatoms\n",
            "thermo_modify flush yes\n",
            "unfix f1\n",
            "unfix f2\n",
        ]

        self.dict["GB_md"] = [
            "# ---------- Run GB MD ---------\n",
            "thermo 100\n",
            f"velocity gbregion create {temp:d} {temp:d}\n",
            f"fix 1 gbregion nvt temp {temp:d} {temp:d} 0.05\n",
            "timestep 0.001\n",
            "run 5000\n",
            "unfix 1\n",
        ]


if __name__ == "__main__":
    from ase.build import bulk

    lammps_name = ""
    comm = None
    log_file = "lammps.log"
    cmd_args = ["-echo", "log", "-log", log_file, "-screen", "none", "-nocite"]
    lmp = lammps(lammps_name, cmd_args, comm)

    struc = bulk("Si", "diamond", cubic=True)
    struc.set_tags([1] * len(struc))
    parameters = [
        "mass * 1.0",
        "pair_style tersoff",
        "pair_coeff * * Si.tersoff Si",
        "neighbor 1.0 bin",
        "neigh_modify  every 1  delay 1  check yes",
    ]
    lammps = LAMMPSlib(lmp=lmp, lmpcmds=parameters)
    struc.set_calculator(lammps)
    print("positions: ")
    print(struc.get_positions())
    print("energy: ")
    print(struc.get_potential_energy())
    print("force: ")
    print(struc.get_forces())
    print("stress: ")
    print(struc.get_stress())

    # dict = LAMMPS_collections().dict
    # for key in dict.keys():
    #    for string in dict[key]:
    #        print(string)



# python -m unittest pyxtal/test_all.py
import unittest
import numpy as np
from pyxtal.lego.SO3 import SO3
from pyxtal import pyxtal
from ase import Atoms
from ase.build import bulk, sort

def calculate_S(atoms, P_ref):
    P = calculator.compute_p(atoms)
    S = np.sum((P - P_ref)**2)
    return S


def numerical_dSdx(x, xtal,  P_ref, eps=1e-4):
    if type(x) == list: x = np.array(x)

    xtal.update_from_1d_rep(x)
    atoms = xtal.to_ase() #* 2
    S0 = calculate_S(atoms, P_ref)
    dSdx = np.zeros(len(x))
    for i in range(len(x)):
        x0 = x.copy()
        x0[i] += eps
        xtal.update_from_1d_rep(x0)
        atoms = xtal.to_ase() #* 2
        S1 = calculate_S(atoms, P_ref)

        x0 = x.copy()
        x0[i] -= eps
        xtal.update_from_1d_rep(x0)
        atoms = xtal.to_ase() #* 2
        S2 = calculate_S(atoms, P_ref)

        dSdx[i] = 0.5*(S1-S2)/eps
    return dSdx


def calculate_dSdx_supercell(x, xtal, P_ref, eps=1e-4):

    xtal.update_from_1d_rep(x)
    atoms = xtal.to_ase() #* 2

    dPdr, P = calculator.compute_dpdr_5d(atoms)

    # Compute dSdr [N, M] [N, N, M, 3, 27] => [N, 3, 27]
    dSdr = np.einsum("ik, ijklm -> jlm", 2*(P - P_ref), dPdr)

    # Get supercell positions
    ref_pos = np.repeat(atoms.positions[:, :, np.newaxis], 27, axis=2)
    for cell in range(27):
        x1, y1, z1 = cell // 9 - 1, (cell // 3) % 3 - 1, cell % 3 - 1
        ref_pos[:, :, cell] += np.array([x1, y1, z1]) @ atoms.cell

    # Compute drdx via numerical func
    drdx = np.zeros([len(atoms), 3, 27, len(x)])

    xtal0 = xtal.copy()
    for i in range(len(x)):
        x0 = x.copy()
        x0[i] += eps
        xtal0.update_from_1d_rep(x0)
        atoms = xtal0.to_ase()

        # Get supercell positions
        pos = np.repeat(atoms.positions[:, :, np.newaxis], 27, axis=2)
        for cell in range(27):
            x1, y1, z1 = cell // 9 - 1, (cell // 3) % 3 - 1, cell % 3 - 1
            pos[:, :, cell] += np.array([x1, y1, z1]) @ atoms.cell

        drdx[:, :, :, i] += (pos - ref_pos)/eps

    # [N, 3, 27] [N, 3, 27, H] => H
    dSdx = np.einsum("ijk, ijkl -> l", dSdr, drdx)
    return dSdx


def get_rotated_cluster(struc, angle=0, axis='x'):
    s_new = struc.copy()
    s_new.rotate(angle, axis)
    cell = 17.22*np.eye(3)
    p_struc = Atoms(s_new.symbols.numbers, positions=s_new.positions, cell=cell, pbc=True)
    return p_struc

def get_perturbed_cluster(struc, p0, p1, eps):
    s_new = struc.copy()
    pos = s_new.positions
    pos[p0, p1] += eps
    cell = 17.22*np.eye(3)
    p_struc = Atoms(s_new.symbols.numbers, positions=pos, cell=cell, pbc=True)
    return p_struc

def get_perturbed_xtal(struc, p0, p1, eps):
    pos = struc.positions.copy()
    pos[p0, p1] += eps
    p_struc = struc.copy()
    p_struc.set_positions(pos)
    return p_struc

def get_dPdR_xtal(xtal, eps):
    p0 = calculator.calculate(xtal, derivative=True)
    shp = p0['x'].shape
    array1 = p0['dxdr']

    for j in range(shp[0]):
        for k in range(3):
            struc = get_perturbed_xtal(xtal, j, k, eps)
            p1 = calculator.calculate(struc)
            array2 = (p1['x'] - p0['x'])/eps
            #if np.linalg.norm(array2) > 1e-2: print(j, k, array2)
            if not np.allclose(array1[:, j, :, k], array2, atol=1e-4):
                print('xtal', j, k, '\n', array1[:, j, :, k], '\n', array2)
            assert(np.allclose(array1[:, j, :, k], array2, atol=1e-4))

# Descriptors Parameters
eps = 1e-8
rc1 = 1.8
rc2 = 3.5
nmax, lmax = 2, 2
calculator = SO3(nmax=nmax, lmax=lmax, rcut=rc1)
calculator0 = SO3(nmax=nmax, lmax=lmax, rcut=rc2)

# NaCl cluster
cluster = bulk('NaCl', crystalstructure='rocksalt', a=5.691694, cubic=True)
cluster = sort(cluster, tags=[0, 4, 1, 5, 2, 6, 3, 7])
cluster.set_pbc((0,0,0))
cluster = get_rotated_cluster(cluster, angle=0.1) # Must rotate

xtal = pyxtal()
xtal.from_prototype('graphite')
atoms = xtal.to_ase()
P_ref = calculator.compute_p(atoms)[0]

# Diamond
class TestCluster(unittest.TestCase):
    struc = get_rotated_cluster(cluster)
    p0 = calculator0.calculate(struc, derivative=True)
    struc = get_rotated_cluster(cluster, 10, 'x')
    p1 = calculator0.calculate(struc)

    def test_SO3_rotation_variance(self):
        array1 = self.p0['x']
        array2 = self.p1['x']
        assert(np.allclose(array1, array2))

    def test_dPdR_vs_numerical(self):
        shp = self.p0['x'].shape
        array1 = self.p0['dxdr']

        for j in range(shp[0]):
            for k in range(3):
                struc = get_perturbed_cluster(cluster, j, k, eps)
                p2 = calculator0.calculate(struc)
                array2 = (p2['x'] - self.p0['x'])/eps
                assert(np.allclose(array1[:,j,:,k], array2, atol=1e-3))

class TestXtal(unittest.TestCase):

    def test_dPdR_diamond(self):
        c = pyxtal()
        c.from_prototype('diamond')
        get_dPdR_xtal(c.to_ase(), eps)

    def test_dPdR_graphite(self):
        c = pyxtal()
        c.from_prototype('graphite')
        get_dPdR_xtal(c.to_ase(), eps)

    def test_dPdR_random(self):
        x = [ 7.952, 2.606, 0.592, 0.926, 0.608, 0.307]
        c = pyxtal()
        c.from_spg_wps_rep(179, ['6a', '6a', '6a', '6a'], x)
        get_dPdR_xtal(c.to_ase(), eps)

    def test_dPdR_random_P(self):
        x = [ 7.952, 2.606, 0.592, 0.926, 0.608, 0.307]
        c = pyxtal()
        c.from_spg_wps_rep(179, ['6a', '6a', '6a', '6a'], x)
        atoms = c.to_ase()
        p0 = calculator.compute_p(atoms)
        _, p1 = calculator.compute_dpdr(atoms)
        _, p2 = calculator.compute_dpdr_5d(atoms)
        assert(np.allclose(p0, p1, atol=1e-3))
        assert(np.allclose(p0, p2, atol=1e-3))

class TestSimilarity(unittest.TestCase):

    def test_sim_diamond(self):
        x = [3.0]
        c = pyxtal()
        c.from_spg_wps_rep(227, ['8a'], x, ['C'])
        atoms = c.to_ase()
        x = c.get_1d_rep_x()
        dSdx1 = numerical_dSdx(x, c, P_ref)
        dSdx2 = calculate_dSdx_supercell(x, c, P_ref)
        #print(dSdx1, dSdx2)
        assert(np.allclose(dSdx1, dSdx2, rtol=1e-1, atol=1e+1))

    def test_dPdR_random(self):
        #x = [ 7.952, 2.606, 0.592, 0.926, 0.608, 0.307]
        x = [9.55,  2.60,  0.48,  0.88,  0.76,   0.36]
        c = pyxtal()
        c.from_spg_wps_rep(179, ['6a', '6a', '6a', '6a'], x)
        atoms = c.to_ase()
        dSdx1 = numerical_dSdx(x, c, P_ref)
        dSdx2 = calculate_dSdx_supercell(x, c, P_ref)
        #print(dSdx1, dSdx2)
        assert(np.allclose(dSdx1, dSdx2, rtol=1e-1, atol=1e+1))

if __name__ == "__main__":
    unittest.main()



from juliacall import Main as jl
# Import the Julia package manager
jl.seval("using CrystalNets")
print("Success CrystalNets")

data = """
12,4.6997,7.1199,3.6592,1.5708,2.1366,1.5708,0,0.438,0.8075,0.6503,2,0.0,0.9344,0.5,-1,-1.0,-1.0,-1.0,-1,-1.0,-1.0,-1.0,-1,-1.0,-1.0,-1.0,-1,-1.0,-1.0,-1.0,-1,-1.0,-1.0,-1.0,-1,-1.0,-1.0,-1.0
43,10.1657,9.213,8.7536,1.5708,1.5708,1.5708,0,0.8343,0.3453,0.1644,0,0.3345,0.7476,0.8868,0,0.7693,0.9651,0.6999,0,0.8732,0.0387,0.0689,0,0.7571,0.9465,0.4255,0,0.9706,0.7866,0.5912,-1,-1.0,-1.0,-1.0,-1,-1.0,-1.0,-1.0
"""

import numpy as np
from pyxtal import pyxtal
from pyxtal.lego.builder import mof_builder

lines = data.strip().splitlines()
dat = [list(map(float, line.split(','))) for line in lines]

xtal = pyxtal(); xtal.from_prototype('graphite')
cif_file = xtal.to_pymatgen()

builder = mof_builder(['C'], [1], verbose=True)
builder.set_descriptor_calculator(mykwargs={'rcut': 2.0})
builder.set_reference_enviroments(cif_file)
builder.set_criteria(CN={'C': [3]})
print(builder)


xtals = []
for i in range(len(dat)):
    rep = dat[i]
    print(i, rep)
    if (len(rep)-7)%4 == 1: rep = rep[1:]
    xtal = pyxtal()
    xtal.from_tabular_representation(rep, normalize=False)
    if xtal.valid and len(xtal.atom_sites) > 0:
        xtals.append(xtal)

xtals_opt = builder.optimize_xtals(xtals,
                                   minimizers=[('Nelder-Mead', 100),
                                               ('L-BFGS-B', 400)],
                                   early_quit=0.15)
xtals = []
builder.db.update_row_topology(overwrite=True, prefix='tmp')
builder.db.clean_structures_spg_topology(dim=3)



# python -m unittest pyxtal/test_all.py
import importlib.util
import os
import unittest

import numpy as np

from pyxtal.lattice import Lattice
from pyxtal.symmetry import Group, Wyckoff_position, get_wyckoffs
from pyxtal.wyckoff_site import atom_site


def resource_filename(package_name, resource_path):
    package_path = importlib.util.find_spec(package_name).submodule_search_locations[0]
    return os.path.join(package_path, resource_path)


cif_path = resource_filename("pyxtal", "database/cifs/")
l01 = Lattice.from_matrix([[4.08, 0, 0], [0, 9.13, 0], [0, 0, 5.50]])
l02 = Lattice.from_para(4.08, 9.13, 5.50, 90, 90, 90)
wp1 = Wyckoff_position.from_group_and_index(36, 0)
wp2 = Wyckoff_position.from_group_and_letter(36, "4a")


class TestWP(unittest.TestCase):
    def test_wp_check_translation(self):
        pass

    def test_wp_site_symm(self):
        data = [
            (143, 1, "3.."),
            (150, 1, ".2."),
            (152, 1, ".2."),
            (154, 1, ".2."),
            (155, 1, ".2"),
            (160, 1, ".m"),
            (160, 2, "3m"),
            (164, 4, ".2/m."),
            (165, 1, ".2."),
            (177, 3, ".2."),
            (178, 2, ".2."),
            (180, 3, ".2."),
            (181, 3, ".2."),
            (230, 6, ".32"),
        ]
        for d in data:
            (sg, i, symbol) = d
            wp = Group(sg)[i]
            wp.get_site_symmetry()
            if wp.site_symm != symbol: print("\n========", wp.site_symm, d, "==========\n")
            assert wp.site_symm == symbol

    def test_wp_dof(self):
        for sg in range(1, 231):
            g = Group(sg)
            for wp in g:
                axs = wp.get_frozen_axis()
                assert wp.get_dof() + len(axs) == 3

    def test_wp_label(self):
        symbol = wp1.get_label()
        assert symbol == "8b"
        symbol = wp2.get_label()
        assert symbol == "4a"

    def test_merge(self):
        pt, wp, _ = wp1.merge([0.05, 0.7, 0.24], l01.matrix, 0.5)
        symbol = wp.get_label()
        assert symbol == "4a"
        pt, wp, _ = wp1.merge([0.15, 0.7, 0.24], l01.matrix, 0.5)
        symbol = wp.get_label()
        assert symbol == "8b"

        wp = Group(167)[0]
        cell = np.diag([9, 9, 7])
        for pt in [[0.12, 0, 0.25], [0, 0.1316, 0.25]]:
            _, wpt, _ = wp.merge(pt, cell, 0.1)
            symbol = wpt.get_label()
            assert symbol == "18e"

        for pt in [[0, 0, 3 / 4], [2 / 3, 1 / 3, 7 / 12]]:
            _, wpt, _ = wp.merge(pt, cell, 0.1)
            symbol = wpt.get_label()
            assert symbol == "6a"

    def test_search_generator(self):
        wp = Group(167)[1]
        for pt in [[0, 0.13, 0.25], [0.13, 0, 0.25]]:
            wp0 = wp.search_generator(pt)
            assert wp0 is not None

    def test_get_wyckoff(self):
        for i in [1, 2, 229, 230]:
            get_wyckoffs(i)
            get_wyckoffs(i, organized=True)

    def test_setting(self):
        wp = Group(14)[0]
        wp.transform_from_matrix()
        assert not wp.is_standard_setting()

    def test_standarize(self):
        pass

    def test_is_equivalent(self):
        g = Group(15)
        wp = g[0]
        a = [0.10052793, 0.12726851, 0.27405404]
        b = [-0.10052642, -0.12726848, -0.27405526]
        c = [0.60052642, 0.62726848, 0.27405526]
        d = [-0.60052642, -0.62726848, -0.27405526]
        e = [0, 2.54537267e-01, 0]
        assert wp.are_equivalent_pts(a, b)
        assert wp.are_equivalent_pts(b, c)
        assert wp.are_equivalent_pts(d, a)
        assert not wp.are_equivalent_pts(a, e)

        wp = g[1]
        a = [0.00, 0.127, 0.254]
        b = [-0.01, -0.127, -0.250]
        assert wp.are_equivalent_pts(a, b)

    def test_project(self):
        pt = np.array([0.0629, 0.1258, 0.25])
        g = Group(178)
        wp = g[1]
        xyz0 = wp.project(pt, np.eye(3))
        diff = np.sum((pt - xyz0) ** 2)
        assert diff < 1e-08

    def test_euclidean(self):
        def check_error(spg, pt, cell):
            p0 = np.dot(pt, cell.matrix)
            for sg in spg:
                wp = Group(sg)[0]
                for i in range(len(wp)):
                    op0 = wp[i]
                    p1 = op0.operate(pt)

                    op1 = wp.get_euclidean_generator(cell.matrix, i)
                    if wp.euclidean:
                        p2 = np.dot(op1.operate(p0), cell.inv_matrix)
                    else:
                        p2 = np.dot(op1.apply_rotation_only(p0), cell.inv_matrix)
                        p2 += op1.translation_vector

                    diff = p1 - p2
                    diff -= np.rint(diff)
                    if np.linalg.norm(diff) > 0.02:
                        # res = '{:2d} {:28s}'.format(i, op0.as_xyz_str())
                        # res += ' {:28s}'.format(op1.as_xyz_str())
                        # res += '{:6.3f} {:6.3f} {:6.3f} -> '.format(*p1)
                        # res += '{:6.3f} {:6.3f} {:6.3f} -> '.format(*p2)
                        # res += '{:6.3f} {:6.3f} {:6.3f}'.format(*diff)
                        # print(res)
                        return False
            return True

        pt = [0.1333, 0.1496, 0.969]

        cell = Lattice.from_para(9.395, 7.395, 8.350, 91, 101, 92, ltype="triclinic")
        assert check_error(range(1, 3), pt, cell)

        cell = Lattice.from_para(9.395, 7.395, 8.350, 90, 101, 90, ltype="monoclinic")
        assert check_error(range(3, 16), pt, cell)

        cell = Lattice.from_para(9.395, 7.395, 8.350, 90, 90, 90, ltype="orthorhombic")
        assert check_error(range(16, 74), pt, cell)

        cell = Lattice.from_para(9.395, 9.395, 8.350, 90, 90, 90, ltype="tetragonal")
        assert check_error(range(74, 143), pt, cell)

        cell = Lattice.from_para(9.395, 9.395, 8.350, 90, 90, 120, ltype="hexagonal")
        assert check_error(range(143, 195), pt, cell)

        cell = Lattice.from_para(9.395, 9.395, 9.395, 90, 90, 90, ltype="cubic")
        assert check_error(range(195, 231), pt, cell)


class Test_wyckoff_site(unittest.TestCase):
    def test_atom_site(self):
        """
        Test the search function
        """
        wp = Group(227)[-5]
        arr = np.array([0.1, 0.1, 0.1])
        for xyz in [
            [0.6501, 0.15001, 0.5999],
            [0.6501, 0.14999, 0.5999],
            [0.6499, 0.14999, 0.5999],
            [0.6499, 0.14999, 0.60001],
        ]:
            site = atom_site(wp, xyz, search=True)
            assert np.allclose(site.position, arr, rtol=0.001)

if __name__ == "__main__":
    unittest.main()



import logging
from time import time

import numpy as np
from pymatgen.core.lattice import Lattice

from pyxtal.operations import *

logging.basicConfig(filename="test.log", level=logging.DEBUG)
eV2GPa = 160.217
GPa2eV = 1.0 / eV2GPa

"""
Scripts to compute LJ energy and force
"""


def get_neighbors(struc, i, rcut):
    """
    script to get the neighbors info for an atom in the struct
    """
    lat = struc.lattice_matrix
    center = struc.cart_coords[i]
    struc.frac_coords[i]
    fcoords = struc.frac_coords
    r, dists, inds, images = Lattice(lat).get_points_in_sphere(fcoords, center=center, r=rcut, zip_results=False)
    ids = np.where(dists > 0.1)
    r = r[ids]
    r = np.dot(r, lat) - center
    return dists[ids] ** 2, r


class LJ:
    """
    LJ model for 3D crystals, maybe extended to 0D, 1D, 2D later
    """

    def __init__(self, epsilon=1.0, sigma=1.0, rcut=8.0):
        """
        passing the parameter to LJ model
        - epsilon
        - sigma
        - rcut
        """

        self.epsilon = epsilon
        self.sigma = sigma
        self.rcut = rcut

    def calc(self, struc, press=1e-4):
        pos = struc.cart_coords
        lat = struc.lattice_matrix
        volume = np.linalg.det(lat)

        # initiate the energy, force, stress
        energy = 0
        force = np.zeros([len(pos), 3])
        stress = np.zeros([3, 3])
        sigma6 = self.sigma**6
        sigma12 = sigma6 * sigma6

        # calculating the energy/force, needs to update sigma/epsilons
        for i, _pos0 in enumerate(pos):
            [r2, r] = get_neighbors(struc, i, self.rcut)
            r6 = np.power(r2, 3)
            r12 = np.power(r6, 2)
            energy += np.sum(4.0 * self.epsilon * (sigma12 / r12 - sigma6 / r6))
            f = (24 * self.epsilon * (2.0 * sigma12 / r12 - sigma6 / r6) / r2)[:, np.newaxis] * r
            force[i] = f.sum(axis=0)
            stress += np.dot(f.T, r)

        energy = 0.5 * energy
        enthalpy = energy + press * volume * GPa2eV
        force = force
        stress = -0.5 * stress / volume * eV2GPa

        return energy, enthalpy, force, stress


class FIRE:
    """
    FIRE algorithm for structure optimization
    """

    def __init__(
        self,
        struc,
        model,
        symmetrize=False,
        e_tol=1e-5,
        f_tol=1e-2,
        dt=0.1,
        maxmove=0.1,
        dtmax=1.0,
        Nmin=10,
        finc=1.1,
        fdec=0.5,
        astart=0.1,
        fa=0.99,
        a=0.1,
    ):
        """
        Parameters for FIRE:
        """

        self.dt = dt
        self.maxmove = maxmove
        self.symmetrize = symmetrize
        self.f_tol = f_tol
        self.e_tol = e_tol
        self.dtmax = dtmax
        self.Nmin = Nmin
        self.finc = finc
        self.fdec = fdec
        self.astart = astart
        self.fa = fa
        self.a = a
        self.nsteps = 0
        self.Nsteps = 0
        self.time0 = time()
        self.model = model
        self.struc = struc
        self.initialize()

    def initialize(self):
        """
        initilaize the positions, energy, force for the 1st step
        """
        self.energy, self.enthalpy, self.force, self.stress = self.model.calc(self.struc)
        self.fmax = np.max(np.abs(np.vstack((self.stress, self.force)).flatten()))
        self.v = np.zeros((len(self.force) + 3, 3))
        # print('Initial Energy: {:12.4f}'.format(self.energy))

    def update(self, freq=50):
        self.energy, self.enthalpy, self.force, self.stress = self.model.calc(self.struc)
        self.fmax = np.max(np.abs(np.vstack((self.stress, self.force)).flatten()))
        if self.nsteps % freq == 0:
            logging.debug(
                f"Step {self.nsteps:4d} Enthalpy: {self.enthalpy:12.4f} Fmax: {self.fmax:12.4f} Vol: {self.volume:6.2f}"
            )

    def step(self):
        f = np.vstack((self.stress, self.force))
        pos = np.vstack((self.struc.lattice_matrix, self.struc.cart_coords))

        vf = np.vdot(f, self.v)
        if vf > 0.0:
            self.v = (1.0 - self.a) * self.v + self.a * f / np.sqrt(np.vdot(f, f)) * np.sqrt(np.vdot(self.v, self.v))
            if self.Nsteps > self.Nmin:
                self.dt = min(self.dt * self.finc, self.dtmax)
                self.a *= self.fa
            self.Nsteps += 1
        else:
            self.v[:] *= 0.0
            self.a = self.astart
            self.dt *= self.fdec
            self.Nsteps = 0

        self.v += self.dt * f
        dr = self.dt * self.v  # needs to impose constraints
        normdr = np.sqrt(np.vdot(dr, dr))
        if normdr > self.maxmove:
            dr = self.maxmove * dr / normdr
        # print('frac0', np.dot(pos[3:,:], np.linalg.inv(pos[:3,:])))

        # Symmetrize the force
        if self.symmetrize:
            # f[:3, :] is the gradient for force, need to symmetrize it as well
            dr[:3, :] = self.symmetrized_stress(dr[:3, :])
            dr[3:, :] = np.dot(self.struc.frac_coords, dr[:3, :])
            f_frac = np.dot(dr[3:, :], np.linalg.inv(pos[:3, :] - dr[:3, :]))
            f_frac = self.symmetrized_force(f_frac)
            dr[3:, :] += np.dot(f_frac, pos[:3, :] - dr[:3, :])

        # pos = pos - dr
        # Symmetrize positions
        # if self.symmetrize:
        #    pos = self.symmetrized_coords(pos)

        # print(self.force)
        # print(self.v)
        self.struc.lattice_matrix = pos[:3, :] - dr[:3, :]
        self.struc.cart_coords = pos[3:, :] - dr[3:, :]
        self.struc.frac_coords = np.dot(self.struc.cart_coords, np.linalg.inv(self.struc.lattice_matrix))
        self.volume = np.linalg.det(self.struc.lattice_matrix)

        # sg = get_symmetry_dataset((pos[:3, :], self.struc.frac_coords, [6]*4), symprec=0.1)['number']
        # print(sg)
        # if self.symmetrize and sg <19:
        #    print(sg)
        #    print('dr\n', dr)
        #    print('pos\n', pos)
        #    print('frac\n', self.struc.frac_coords)
        #    print('force\n', f_frac)
        #    sys.exit()
        self.update()

    def run(self, max_steps=1000):
        self.max_steps = max_steps
        while self.nsteps < max_steps:
            if self.check_convergence():
                break
            self.step()
            self.nsteps += 1

        logging.info(
            f"Finish at Step {self.nsteps:4d} Enthalpy: {self.enthalpy:12.4f} Fmax: {self.fmax:12.4f} Time: {time() - self.time0:6.2f} seconds"
        )

    def check_convergence(self):
        """
        There should be two options to terminate the optimization
        before it reaches the maximum number of cycles
        1, by energy difference compared to the previous step: self.e_tol
        2, by the residual force: self.f_tol
        Will implement both options later.
        """
        converged = False
        if self.fmax < self.f_tol and self.nsteps > 0:
            converged = True
        return converged

    def symmetrized_coords(self, coords):
        start_index = 0
        new_coords = []
        for ws in self.struc.wyckoff_sites:
            # Get coordinates associated with WP
            original_coords = coords[start_index : start_index + ws.multiplicity]
            # Re-generate the points from the first generating point
            gen_coord = coords[start_index]
            wp_coords0 = apply_ops(gen_coord, ws.wp.generators_m)
            # Calculate the PBC translation applied to each point
            translations = original_coords - wp_coords0
            frac_translations = np.dot(translations, np.linalg.inv(self.struc.lattice_matrix))
            frac_translations = np.round(frac_translations)
            cart_translations = np.dot(frac_translations, self.struc.lattice_matrix)
            # Subtract translations from original coords
            translated_coords = original_coords - cart_translations
            # Apply inverse WP operations to get generating_points
            inverse_ops = ws.wp.inverse_generators_m
            generating_points = apply_ops_diagonal(translated_coords, inverse_ops)
            # Find the average of the generating points
            average_point = np.sum(generating_points, axis=0) / ws.multiplicity
            # Convert to fractional coordinates
            average_point = np.dot(average_point, np.linalg.inv(self.struc.lattice_matrix))
            # Apply the WP operations to the averaged generating point
            wp_coords = apply_ops(average_point, ws.wp)
            # Convert back to Cartesian coordintes
            wp_coords = np.dot(wp_coords, self.struc.lattice_matrix)
            # Re-apply the cartesian translations
            wp_coords = wp_coords + cart_translations
            new_coords = wp_coords if len(new_coords) == 0 else np.vstack([new_coords, wp_coords])
            start_index += ws.multiplicity
        return new_coords

    def symmetrized_force(self, coords):
        start_index = 0
        new_coords = []
        for ws in self.struc.wyckoff_sites:
            # Get coordinates associated with WP
            coords[start_index : start_index + ws.multiplicity]
            # Re-generate the points from the first generating point
            gen_coord = coords[start_index]
            wp_coords0 = apply_ops(gen_coord, ws.wp.generators_m)
            # Apply inverse WP operations to get generating_points
            inverse_ops = ws.wp.inverse_generators_m
            generating_points = apply_ops_diagonal(wp_coords0, inverse_ops)
            # Find the average of the generating points
            average_point = np.sum(generating_points, axis=0) / ws.multiplicity
            # Convert to fractional coordinates
            average_point = np.dot(average_point, np.linalg.inv(self.struc.lattice_matrix))
            # For forces, we do not apply translational WP operations, so we truncate the ops
            matrices = np.array([op.affine_matrix[:3, :3] for op in ws.wp])
            # Apply the truncated WP operations to the averaged generating point
            wp_coords = np.dot(average_point, matrices)
            # Convert back to Cartesian coordintes
            wp_coords = np.dot(wp_coords, self.struc.lattice_matrix)
            new_coords = wp_coords if len(new_coords) == 0 else np.vstack([new_coords, wp_coords])
            start_index += ws.multiplicity
        # if np.sum(np.abs(coords-new_coords))>1e-1:
        #    print(coords)
        #    print(new_coords)
        #    print(coords-new_coords)
        #    sys.exit()
        return new_coords

    def symmetrized_stress(self, stress):
        # Make the matrix lower-diagonal
        for i, j in [(1, 0), (2, 1), (2, 2)]:
            stress[i][j] = stress[i][j] + stress[j][i]
            stress[j][i] = 0
        # Normalize the matrix
        snm = self.struc.lattice.stress_normalization_matrix
        m2 = np.multiply(stress, snm)
        # Normalize the on-diagonal elements
        indices = self.struc.lattice.stress_indices
        if len(indices) == 2:
            total = 0
            for index in indices:
                total += stress[index]
            value = np.sqrt(total)
            for index in inices:
                m2[index] = value
        elif len(indices) == 3:
            total = 0
            for index in indices:
                total += stress[index]
            value = np.cbrt(total)
            for index in inices:
                m2[index] = value
        return m2


if __name__ == "__main__":
    from spglib import get_symmetry_dataset
    from pyxtal import pyxtal

    for _i in range(10):
        crystal = pyxtal()
        crystal.from_random(3, 11, ["C"], [4])
        if crystal.valid:
            crystal1 = deepcopy(crystal)
            test = LJ(epsilon=0.01, sigma=3.40, rcut=8.0)
            struc = (crystal1.lattice_matrix, crystal1.frac_coords, [6] * 4)
            eng, enth, force, stress = test.calc(crystal1)
            sg = get_symmetry_dataset(struc)["number"]
            print(f"\nBefore relaxation Space group:            {sg:4d}   Energy: {eng:12.4}  Enthalpy: {enth:12.4}")

            dyn1 = FIRE(crystal1, test, f_tol=1e-5, dt=0.2, maxmove=0.2)  # , symmetrize=True)
            dyn1.run(500)
            eng, enth, force, stress = test.calc(crystal1)
            struc = (dyn1.struc.lattice_matrix, dyn1.struc.frac_coords, [6] * 4)
            sg = get_symmetry_dataset(struc, symprec=0.1)["number"]
            print(f"After relaxation without symm Space group: {sg:4d}  Energy: {eng:12.4}  Enthalpy: {enth:12.4}")

            dyn1 = FIRE(crystal, test, f_tol=1e-5, dt=0.2, maxmove=0.2, symmetrize=True)
            dyn1.run(500)
            eng, enth, force, stress = test.calc(crystal)
            struc = (dyn1.struc.lattice_matrix, dyn1.struc.frac_coords, [6] * 4)
            sg = get_symmetry_dataset(struc, symprec=0.02)["number"]
            if sg is None:
                sg = 0
            print(f"After relaxation with symm Space group:    {sg:4d}  Energy: {eng:12.4}  Enthalpy: {enth:12.4}")
            # dyn1 = FIRE(crystal, test, f_tol=1e-5, dt=0.2, maxmove=0.2)
            # struc = (dyn1.struc.lattice_matrix, dyn1.struc.frac_coords, [6]*4)
            # sg =  get_symmetry_dataset(struc, symprec=0.02)['number']
            # print('After relaxation without symm Space group: {:4d}  Energy: {:12.4}  Enthalpy: {:12.4}'.format(sg, eng, enth))
            # right now, it seems structures goes to either HCP of FCC after relaxation, which is expected for 3D LJ system
            # need to compare with other code to see if the energy is correct



import psutil
import os

def print_memory_usage():
    process = psutil.Process(os.getpid())
    mem_info = process.memory_info()
    print(f"Memory usage: {mem_info.rss / 1024 ** 2:.2f} MB")  # RSS is the Resident Set Size

if __name__ == "__main__":
    print_memory_usage()
    from pyxtal.symmetry import Group

    print("\nImport group")
    print_memory_usage()

    print("\nCreate layer Group with quick")
    g = Group(4, dim=2, quick=True)
    print_memory_usage()

    print("\nCreate Group with quick")
    g = Group(4, quick=True)
    print_memory_usage()

    print("\nCreate Group")
    g = Group(227)
    print_memory_usage()

    print("\nCall pyxtal")
    from pyxtal import pyxtal
    xtal = pyxtal()
    xtal.from_spg_wps_rep(194, ['2c', '2b'], [2.46, 6.70])
    print_memory_usage()

    print("\nCall subgroup")
    xtal.subgroup_once()
    print_memory_usage()



# python -m unittest pyxtal/test_all.py
import importlib.util
import os
import unittest

import numpy as np
import pymatgen.analysis.structure_matcher as sm
from pymatgen.core import Structure

from pyxtal import pyxtal
from pyxtal.lattice import Lattice
from pyxtal.symmetry import Group, Hall, Wyckoff_position


def resource_filename(package_name, resource_path):
    package_path = importlib.util.find_spec(package_name).submodule_search_locations[0]
    return os.path.join(package_path, resource_path)


cif_path = resource_filename("pyxtal", "database/cifs/")
l01 = Lattice.from_matrix([[4.08, 0, 0], [0, 9.13, 0], [0, 0, 5.50]])
l02 = Lattice.from_para(4.08, 9.13, 5.50, 90, 90, 90)
wp1 = Wyckoff_position.from_group_and_index(36, 0)
wp2 = Wyckoff_position.from_group_and_letter(36, "4a")


class TestLattice(unittest.TestCase):
    def test_para_matrix(self):
        assert np.allclose(l01.matrix, l02.matrix)

    def test_swap(self):
        l01.swap_axis(ids=[1, 0, 2])
        abc = l01.get_para()[:3]
        assert abc, np.array([9.13, 4.08, 5.5])

    def test_optimize_once(self):
        l3 = Lattice.from_para(4.08, 7.13, 5.50, 90, 38, 90, ltype="monoclinic")
        lat, tran, _ = l3.optimize_once()
        assert abs(lat.beta - 1.495907) < 0.0001

    def test_optimize_multi(self):
        l4 = Lattice.from_para(71.364, 9.127, 10.075, 90.00, 20.80, 90.00, ltype="monoclinic")
        lat, _ = l4.optimize_multi(7)
        assert abs(lat.beta - 1.7201) < 0.01

    def test_setpara(self):
        l0 = Lattice.from_matrix([[4.08, 0, 0], [0, 9.13, 0], [0, 0, 5.50]])
        l0.set_para([5, 5, 5, 90, 90, 90])
        assert l0.a == 5

    def test_search_transformation(self):
        l6 = Lattice.from_para(3.454, 3.401, 5.908, 90.00, 105.80, 90.00, ltype="monoclinic")
        l7 = Lattice.from_para(6.028, 3.419, 6.028, 90.00, 146.92, 90.00, ltype="monoclinic")
        l7, _ = l7.optimize_multi()
        trans, diff = l7.search_transformation(l6)
        l7 = l7.transform_multi(trans)
        assert np.abs(l7.matrix - l6.matrix).sum() < 0.25

    def test_is_valid_lattice(self):
        l8 = Lattice.from_para(3.454, 3.401, 5.908, 90.00, 105.80, 91.00, ltype="monoclinic")
        l9 = Lattice.from_para(3.454, 3.401, 5.908, 90.00, 105.80, 90.00, ltype="monoclinic")
        l10 = Lattice.from_para(3.454, 3.401, 5.908, 90.00, 90.00, 90.00, ltype="cubic")
        assert not l8.is_valid_lattice()
        assert l9.is_valid_lattice()
        assert not l10.is_valid_lattice()

    def test_from_1d_representation(self):
        lat = Lattice.from_1d_representation([5.09, 6.11], "trigonal")
        assert abs(lat.a - 5.09) < 0.001
        assert abs(lat.c - 6.11) < 0.001
        assert abs(lat.gamma - 2 / 3 * np.pi) < 0.001


class TestOptimizeLattice(unittest.TestCase):
    def test_atomic(self):
        c1 = pyxtal()
        c1.from_seed(cif_path + "LiCs.cif", backend="pyxtal")
        pmg1 = c1.to_pymatgen()

        c2 = c1.copy()
        c2.optimize_lattice(1)
        pmg2 = c2.to_pymatgen()
        assert sm.StructureMatcher().fit(pmg1, pmg2)

        c2.optimize_lattice(1)
        pmg2 = c2.to_pymatgen()
        assert sm.StructureMatcher().fit(pmg1, pmg2)

        c3 = pyxtal()
        c3.from_seed(cif_path + "LiCs.cif")
        pmg3 = c3.to_pymatgen()
        assert sm.StructureMatcher().fit(pmg1, pmg3)

    def test_molecular_from_seed(self):
        c1 = pyxtal(molecular=True)
        c1.from_seed(seed=cif_path + "aspirin-c.cif", molecules=["aspirin"])
        c1.mol_sites[0].get_mol_object(0)
        c1.mol_sites[0].get_mol_object(1)
        pmg1 = c1.to_pymatgen()

        c2 = c1.copy()
        c2.optimize_lattice(1)
        pmg2 = c2.to_pymatgen()
        assert sm.StructureMatcher().fit(pmg1, pmg2)

    def test_molecular_trans(self):
        c1 = pyxtal(molecular=True)
        c1.from_seed(seed=cif_path + "aspirin.cif", molecules=["aspirin"])
        pmg1 = c1.to_pymatgen()
        c1.transform(trans=[[1, 0, 0], [0, 1, 0], [-1, 0, 1]])
        pmg2 = c1.to_pymatgen()
        c1.optimize_lattice()
        pmg3 = c1.to_pymatgen()
        assert sm.StructureMatcher().fit(pmg1, pmg2)
        assert sm.StructureMatcher().fit(pmg2, pmg3)

    def test_molecular_nonstd(self):
        sgs = [5, 7, 8, 9, 12, 13, 14, 15]
        c1 = pyxtal(molecular=True)
        for sg in sgs:
            hns = Hall(sg).hall_numbers
            for hn in hns[1:]:
                c1.from_random(3, hn, ["aspirin"], use_hall=True)
                pmg1 = c1.to_pymatgen()
                c2 = c1.copy()
                c2.optimize_lattice(1)
                pmg2 = c2.to_pymatgen()
                assert sm.StructureMatcher().fit(pmg1, pmg2)
                pmg3 = Structure.from_str(c1.to_file(), fmt="cif")
                assert sm.StructureMatcher().fit(pmg1, pmg3)

    def test_molecular_from_random(self):
        rng = np.random.default_rng(0)
        sgs = [5, 7, 8, 12, 13, 14]
        c1 = pyxtal(molecular=True)
        for _i in range(20):
            sg = rng.choice(sgs)
            c1.from_random(3, sg, ["aspirin"])
            pmg1 = c1.to_pymatgen()
            c2 = c1.copy()
            c2.optimize_lattice(1)
            pmg2 = c2.to_pymatgen()
            assert sm.StructureMatcher().fit(pmg1, pmg2)

    def test_transform(self):
        l = Lattice.from_para(48.005, 7.320, 35.864, 90.000, 174.948, 90.000, ltype="monoclinic")
        sites = [
            [0.0000, 0.0000, 0.1250],
            [0.2296, 0.7704, 0.5370],
            [0.2296, 0.2296, 0.5370],
            [0.5408, 0.0000, 0.6186],
            [0.0000, 0.0000, 0.3074],
        ]
        c = pyxtal()
        c.build(9, ["S"], [8], lattice=l, sites=[sites])
        pmg0 = c.to_pymatgen()
        c.optimize_lattice()
        pmg1 = c.to_pymatgen()
        assert sm.StructureMatcher().fit(pmg0, pmg1)

    def test_build(self):
        l = np.array([48.005, 7.320, 35.864, 90.000, 174.948, 90.000])
        sites = [
            [0.0000, 0.0000, 0.1250],
            [0.2296, 0.7704, 0.5370],
            [0.2296, 0.2296, 0.5370],
            [0.5408, 0.0000, 0.6186],
            [0.0000, 0.0000, 0.3074],
        ]
        c = pyxtal()
        c.build(9, ["S"], [8], lattice=l, sites=[sites])
        pmg0 = c.to_pymatgen()
        l1 = c.lattice.matrix

        c.optimize_lattice()
        pmg1 = c.to_pymatgen()
        assert sm.StructureMatcher().fit(pmg0, pmg1)

        c1 = pyxtal()
        c1.build(9, ["S"], [8], lattice=l1, sites=[sites])
        c1.optimize_lattice()
        pmg2 = c1.to_pymatgen()
        assert sm.StructureMatcher().fit(pmg0, pmg2)

    def test_build_1D(self):
        lat = Lattice.from_para(6.8472, 6.8472, 3.3198, 90, 90, 90, "tetragonal")
        c1 = pyxtal()
        sites = [
            [
                ("4g", 0.9236547993389047, 0.0, 0.5),
                ("4f", 0.39177300078207977, 0.0, 0.0),
            ]
        ]
        c1.build(30, ["C"], [8], lat, sites=sites, dim=1)
        assert c1.valid

    def test_build_2D(self):
        lat = Lattice.from_para(6.8472, 6.8472, 3.3198, 90, 90, 90, "tetragonal")
        c1 = pyxtal()
        sites = [[("4c", 0.15, 0.50, 0.28), ("4c", 0.59, 0.85, 0.87)]]
        c1.build(30, ["C"], [8], lat, sites=sites, dim=2)
        assert c1.valid

    # def test_build_0D(self):
    #    lat = Lattice.from_para(2.85, 2.85, 2.85, 90, 90, 90, 'spherical')
    #    c1 = pyxtal()
    #    sites = [[('8b', 0.70, 0.70, 0.70)]]
    #    c1.build(30, ['C'], [8], lat, sites=sites, dim=0)
    #    self.assertTrue(c1.valid)

    def test_transforms(self):
        paras = [
            (5.0317, 19.2982, 5.8004, 90.0000, 122.2672, 90.0000, "monoclinic", 6),
            (5.0317, 19.2982, 5.8004, 90.0000, 57.7328, 90.0000, "monoclinic", 6),
            (9.0640, 8.3522, 5.2856, 90.0000, 103.9699, 90.0000, "monoclinic", 3),
            (9.4913, 8.5844, 5.3358, 90.0000, 110.0035, 90.0000, "monoclinic", 3),
            (3.7727, 6.7490, 7.6446, 64.4392, 77.9087, 75.7214, "triclinic", 1),
            (3.7297, 6.5421, 7.9915, 73.9361, 71.5867, 103.2590, "triclinic", 1),
            (5.5740, 7.0902, 11.4529, 96.5703, 90.0224, 108.1857, "triclinic", 2),
            (5.5487, 7.1197, 11.4349, 83.0223, 89.8479, 72.0003, "triclinic", 2),
            (5.7985, 30.6352, 7.6374, 90.0000, 112.2615, 90.0000, "monoclinic", 3),
            (5.8280, 30.5992, 7.6373, 90.0000, 112.6020, 90.0000, "monoclinic", 3),
        ]

        for i in range(int(len(paras) / 2)):
            (a1, b1, c1, alpha1, beta1, gamma1, ltype1, N) = paras[i * 2]
            (a2, b2, c2, alpha2, beta2, gamma2, ltype2, N) = paras[i * 2 + 1]
            l1 = Lattice.from_para(a1, b1, c1, alpha1, beta1, gamma1, ltype=ltype1)
            l2 = Lattice.from_para(a2, b2, c2, alpha2, beta2, gamma2, ltype=ltype2)
            trans, diffs = l2.search_transformations(l1)
            print(i, len(trans), N)
            assert len(trans) == N

            # print("\nTarget:", l1)
            # print("Start :", l2)
            # if len(trans) == N:
            # for tran, diff in zip(trans, diffs):
            #    l = l2.transform_multi(tran)
            #    strs = "Success:" + str(l) + " {:6.3f} {:6.3f} {:6.3f}".format(*diff)
            #    print(strs)

    def test_optlat_setting(self):
        paras = [
            (18.950, 10.914, 31.672, 90, 168.63, 90, "monoclinic"),
            (48.005, 7.320, 35.864, 90, 174.948, 90, "monoclinic"),
        ]
        sites = [0.3600, 0.7500, 0.6743]
        for para in paras:
            (a, b, c, alpha, beta, gamma, ltype) = para
            l = Lattice.from_para(a, b, c, alpha, beta, gamma, ltype=ltype)
            for sg in range(3, 16):
                mult = Group(sg, quick=True)
                c = pyxtal()
                c.build(sg, ["S"], [mult], lattice=l, sites=[[sites]])
                pmg0 = c.to_pymatgen()
                c0 = c.copy()
                c0.optimize_lattice(standard=True)
                pmg1 = c0.to_pymatgen()
                c1 = c.copy()
                c1.optimize_lattice(standard=False)
                pmg2 = c1.to_pymatgen()
                d1 = sm.StructureMatcher().get_rms_dist(pmg0, pmg1)
                d2 = sm.StructureMatcher().get_rms_dist(pmg0, pmg2)
                assert sum(d1) + sum(d2) < 0.001

if __name__ == "__main__":
    unittest.main()



"""
Global Optimizer
"""

from __future__ import annotations

from time import time
from typing import TYPE_CHECKING

import numpy as np
from scipy.stats import qmc

from numpy.random import Generator
from pymatgen.analysis.structure_matcher import StructureMatcher

from pyxtal.optimize.base import GlobalOptimize
from pyxtal.representation import representation
from pyxtal.lattice import Lattice

if TYPE_CHECKING:
    from pyxtal.molecule import pyxtal_molecule

def generate_qrs_cell(sampler, cell_bounds, ref_volume, ltype):
    """
    A routine to generate quasi random samples for lattice and wp
    """
    # Sample cell parameters
    min_vol, max_vol = 0.75*ref_volume, 2.5*ref_volume
    lb = [b[0] for b in cell_bounds]
    ub = [b[1] for b in cell_bounds]
    count = 0
    while True:
        count += 1
        sample = qmc.scale(sampler.random(), lb, ub)[0].tolist()
        lat = Lattice.from_1d_representation(sample, ltype)
        if min_vol < lat.volume < max_vol:
            #print(sample, ltype)
            return sample
        if count == 1000:
            raise ValueError("Cannot generate valid cell with 1000 attempts")

def generate_qrs_xtals(cell, wp_bounds, N_pop, smiles, comp, sampler_wp=None, d_tol=0.85):
    """
    Get the qrs xtal samples

    Args:
        cell (list): [hall, a, b, c]
        wp_bounds (list): [[wp0], [wp1], ...]
        N_pop (int): number of valid candidates
        smiles (list): []
        comp (list): [1]
        sampler_wp: sampler
        d_tol (float): short distance tolerance value
    """
    #cell = [81, 11.38,  6.48, 11.24,  96.9]
    xtals = []
    lb, ub = [], []
    seqs = []
    for wp_bound in wp_bounds:
        lb += [b[0] for b in wp_bound]
        ub += [b[1] for b in wp_bound]
        seqs.append(len(wp_bound))

    if sampler_wp is None:
        sampler_wp = qmc.Sobol(d=len(lb), scramble=False)

    m = max([int(np.log2(N_pop))+3, 9])

    for i in range(2**m):
        sample_wp = sampler_wp.random()#; print(sample_wp)
        sample_wp = qmc.scale(sample_wp, lb, ub)[0].tolist()
        x = [cell]
        prev = 0
        for seq in seqs:
            wp = [0] + sample_wp[prev:prev+seq] + [0]#; print('DDDD', prev, prev+seq, sample_wp[prev:prev+seq], wp)
            x.append(wp) #, [0] + wp0.tolist() + [False]] #print(x)
            prev = seq
        rep = representation(x, smiles)
        xtal = rep.to_pyxtal(composition=comp)
        if not xtal.has_special_site() and len(xtal.check_short_distances(r=d_tol)) == 0:
            #print("debug", rep)
            xtals.append((xtal, "QRandom"))
            if len(xtals) == N_pop:
                return xtals
        #else:
        #    print(rep, len(xtal.check_short_distances(r=0.6)))
    return xtals


class QRS(GlobalOptimize):
    """
    Quasi-Random Sampling

    Args:
        smiles (str): smiles string
        workdir (str): path of working directory
        sg (int or list): space group number or list of spg numbers
        tag (string): job prefix
        ff_opt (bool): activate on the fly FF mode
        ff_style (str): automated force style (`gaff` or `openff`)
        ff_parameters (str or list): ff parameter xml file or list
        reference_file (str): path of reference xml data for FF training
        N_gen (int): number of generation (default: `10`)
        N_pop (int): number of populations (default: `10`)
        N_cpu (int): number of cpus for parallel calculation (default: `1`)
        cif (str): cif file name to store all structure information
        block: block mode
        num_block: list of blocks
        compositions: list of composition, (default is [1]*Num_mol)
        lattice (bool): whether or not supply the lattice
        torsions: list of torsion angle
        molecules (list): list of pyxtal_molecule objects
        sites (list): list of wp sites, e.g., [['4a']]
        use_hall (bool): whether or not use hall number (default: False)
        skip_ani (bool): whether or not use ani or not (default: True)
        eng_cutoff (float): the cutoff energy for FF training
        E_max (float): maximum energy defined as an invalid structure
        verbose (bool): show more details
        use_mpi: bool = False,
    """

    def __init__(
        self,
        smiles: str,
        workdir: str,
        sg: int | list,
        tag: str = 'test',
        info: dict[any, any] | None = None,
        ff_opt: bool = False,
        ff_style: str = "openff",
        ff_parameters: str = "parameters.xml",
        reference_file: str = "references.xml",
        ref_criteria: dict[any, any] | None = None,
        N_gen: int = 10,
        N_pop: int = 10,
        N_cpu: int = 1,
        cif: str | None = None,
        block: list[any] | None = None,
        num_block: list[any] | None = None,
        composition: list[any] | None = None,
        lattice: Lattice | None = None,
        torsions: list[any] | None = None,
        molecules: list[pyxtal_molecule] | None = None,
        sites: list[any] | None = None,
        use_hall: bool = False,
        skip_ani: bool = True,
        factor: float = 1.1,
        eng_cutoff: float = 5.0,
        E_max: float = 1e10,
        N_survival: int = 20,
        verbose: bool = False,
        random_state: int | None = None,
        max_time: float | None = None,
        matcher: StructureMatcher | None = None,
        early_quit: bool = False,
        check_stable: bool = False,
        use_mpi: bool = False,
    ):

        # POPULATION parameters:
        self.N_gen = N_gen # Number of lattice points
        self.N_pop = N_pop # Number of wp varieties
        self.verbose = verbose
        self.name = 'QRS'

        # initialize other base parameters
        GlobalOptimize.__init__(
            self,
            smiles,
            workdir,
            sg,
            tag,
            info,
            ff_opt,
            ff_style,
            ff_parameters,
            reference_file,
            ref_criteria,
            N_cpu,
            cif,
            block,
            num_block,
            composition,
            lattice,
            torsions,
            molecules,
            sites,
            use_hall,
            skip_ani,
            factor,
            eng_cutoff,
            E_max,
            None, #random_state,
            max_time,
            matcher,
            early_quit,
            check_stable,
            use_mpi,
        )

        # Setup the stats [N_gen, Npop, (E, matches)]
        self.stats = np.zeros([self.N_gen, self.N_pop, 2])
        self.stats[:, :, 0] = self.E_max


        if self.rank == 0:
            strs = self.full_str()
            self.logging.info(strs)
            print(strs)

    def full_str(self):
        s = str(self)
        s += "\nMethod    : Deterministic Quasi-Random Sampling"
        s += f"\nGeneration: {self.N_gen:4d}"
        s += f"\nPopulation: {self.N_pop:4d}"
        # The rest base information from now on
        return s

    def _run(self, pool=None):
        """
        The main code to run QRS prediction

        Returns:
            success_rate or None
        """
        self.ref_volumes = []
        N_added = 0
        success_rate = 0
        print(f"Rank {self.rank} starts QRS in {self.tag}")

        for gen in range(self.N_gen):
            self.generation = gen
            cur_xtals = None
            self.logging.info(f"Gen {gen} starts in Rank {self.rank}")

            if self.rank == 0:
                print(f"\nGeneration {gen:d} starts")
                self.logging.info(f"Generation {gen:d} starts")
                t0 = time()

                # Initialize
                cur_xtals = [(None, "Random")] * self.N_pop

                # QRS update
                if gen > 0:
                    if self.lattice is not None:
                        cell = [self.hall_number] + self.lattice.encode()
                        sampler = self.sampler
                    else:
                        cell = generate_qrs_cell(self.sampler,
                                                 self.cell_bounds,
                                                 self.ref_volumes[-1],
                                                 self.ltype)
                        cell = [self.hall_number] + cell
                        sampler = None

                    cur_xtals = generate_qrs_xtals(cell,
                                                   self.wp_bounds,
                                                   self.N_pop,
                                                   self.smiles,
                                                   self.composition,
                                                   sampler)
                    strs = f"Cell parameters in Gen-{gen:d}: "
                    print(strs, cell, self.ref_volumes[-1], len(cur_xtals))


            # Broadcast
            if self.use_mpi:
                cur_xtals = self.comm.bcast(cur_xtals, root=0)

            # Local optimization
            gen_results = self.local_optimization(cur_xtals, qrs=True, pool=pool)
            self.logging.info(f"Rank {self.rank} finishes local_opt")

            # Summary and Ranking
            quit = False
            if self.rank == 0:
                cur_xtals, matches, engs = self.gen_summary(t0,
                                        gen_results, cur_xtals)

                # update hist_best
                vols = []
                for id, (xtal, _) in enumerate(cur_xtals):
                    if xtal is not None:
                        vols.append(xtal.lattice.volume)

                # update best volume
                self.ref_volumes.append(np.array(vols).mean())
                if gen == 0:
                    best_xtal = cur_xtals[0][0]
                    self.cell_bounds = best_xtal.lattice.get_bounds(2.5, 25)
                    self.ltype = best_xtal.lattice.ltype
                    self.wp_bounds = [site.get_bounds() for site in best_xtal.mol_sites]
                    self.hall_number = best_xtal.group.hall_number
                    if self.lattice is not None:
                        len_reps = sum(len(bound) for bound in self.wp_bounds)
                    else:
                        len_reps = len(self.cell_bounds)
                    self.sampler = qmc.Sobol(d=len_reps, scramble=False)

                if self.ref_pmg is not None:
                    success_rate = self.success_count(cur_xtals, matches)

                    if self.early_termination(success_rate):
                        quit = True

                elif self.ref_pxrd is not None:
                    self.count_pxrd_match(cur_xtals, matches)

            if self.use_mpi:
                quit = self.comm.bcast(quit, root=0)
                self.comm.Barrier()
                self.logging.info(f"Gen {gen} Finish in Rank {self.rank}")

            if quit:
                self.logging.info(f"Early Termination in Rank {self.rank}")
                return success_rate

        return success_rate

if __name__ == "__main__":
    print("test")



# python -m unittest pyxtal/test_all.py
import importlib.util
import os
import unittest

from pyxtal import pyxtal
from pyxtal.lattice import Lattice
from pyxtal.symmetry import Wyckoff_position
from pyxtal.XRD import Similarity


def resource_filename(package_name, resource_path):
    package_path = importlib.util.find_spec(package_name).submodule_search_locations[0]
    return os.path.join(package_path, resource_path)


cif_path = resource_filename("pyxtal", "database/cifs/")
l01 = Lattice.from_matrix([[4.08, 0, 0], [0, 9.13, 0], [0, 0, 5.50]])
l02 = Lattice.from_para(4.08, 9.13, 5.50, 90, 90, 90)
wp1 = Wyckoff_position.from_group_and_index(36, 0)
wp2 = Wyckoff_position.from_group_and_letter(36, "4a")


class TestPXRD(unittest.TestCase):
    def test_similarity(self):
        C1 = pyxtal()
        C1.from_random(3, 227, ["C"], [8], sites=[["8a"]])
        xrd1 = C1.get_XRD()
        C2 = C1.subgroup_once(eps=1e-3)
        xrd2 = C1.get_XRD()
        p1 = xrd1.get_profile()
        p2 = xrd2.get_profile()
        s = Similarity(p1, p2, x_range=[15, 90])
        assert 0.9 < s.value < 1.001

        C2.apply_perturbation(1e-3, 1e-3)
        xrd3 = C2.get_XRD()
        xrd3.get_profile()
        s = Similarity(p1, p2, x_range=[15, 90])
        assert 0.95 < s.value < 1.001

if __name__ == "__main__":
    unittest.main()
